from voluptuous import Invalid
import unittest
from bw2data.validate import *


class ValidationTestCase(unittest.TestCase):
    def test_valid_tuple(self):
        with self.assertRaises(Invalid):
            valid_tuple(())
        with self.assertRaises(Invalid):
            valid_tuple(["a", "b"])
        with self.assertRaises(Invalid):
            valid_tuple((1, "b"))
        with self.assertRaises(Invalid):
            valid_tuple(("b", 1))
        self.assertTrue(valid_tuple(("a", "b")))
        with self.assertRaises(Invalid):
            self.assertTrue(valid_tuple(("a", ())))
        with self.assertRaises(Invalid):
            self.assertTrue(valid_tuple(("a", [])))
        self.assertTrue(valid_tuple(("a", "1")))

    def test_uncertainty_dict(self):
        schema = Schema(uncertainty_dict)
        with self.assertRaises(Invalid):
            schema({})
        with self.assertRaises(Invalid):
            schema({"loc": 0})
        with self.assertRaises(Invalid):
            schema({"amount": 0, "foo": "bar"})
        self.assertTrue(schema({"amount": 0}))

    def test_maybe_uncertainty(self):
        schema = Schema(maybe_uncertainty)
        self.assertTrue(schema({"amount": 0}))
        self.assertTrue(schema(4))
        self.assertTrue(schema(4.2))

    def test_exchange(self):
        schema = Schema(exchange)
        with self.assertRaises(Invalid):
            schema({})
        with self.assertRaises(Invalid):
            schema({"amount": 1})
        with self.assertRaises(Invalid):
            schema({"input": ("a", 1), "type": "foo"})
        self.assertTrue(schema({"amount": 1, "input": ("a", "1"), "type": "foo"}))

    def test_db_validator(self):
        self.assertTrue(db_validator({("a", "1"): {}}))
        self.assertTrue(db_validator({("a", "1"): {"type": "foo", "exchanges": [],}}))
        self.assertTrue(db_validator({("a", "1"): {"name": "foo", "exchanges": [],}}))
        self.assertTrue(db_validator({("a", "1"): {"name": "foo", "type": "bar",}}))
        self.assertTrue(
            db_validator({("a", "1"): {"name": "foo", "type": "bar", "exchanges": [],}})
        )
        self.assertTrue(
            db_validator(
                {
                    ("a", "1"): {
                        "name": "foo",
                        "type": "bar",
                        "exchanges": [],
                        "night": "day",
                    }
                }
            )
        )

    def test_ia_validator(self):
        self.assertTrue(ia_validator([[("a", "1"), 2.0]]))
        self.assertTrue(ia_validator([[("a", "1"), 2.0, "CH"]]))
        self.assertTrue(ia_validator([[("a", "1"), 2.0, "CH"], [("a", "1"), 2.0],]))

    def test_weighting_too_long(self):
        with self.assertRaises(Invalid):
            weighting_validator([{"amount": 0}, {"amount", 0}])

    def test_weighting_too_short(self):
        with self.assertRaises(Invalid):
            weighting_validator([])

    def test_weighting(self):
        self.assertTrue(weighting_validator([{"amount": 0}]))



#!/usr/bin/env python
"""Brightway2 updating made simple.

Usage:
  bw2-uptodate
  bw2-uptodate -l | --list
  bw2-uptodate -h | --help
  bw2-uptodate --version

Options:
  --list        List the updates needed, but don't do anything
  -h --help     Show this screen.
  --version     Show version.

"""
from docopt import docopt
import sys
import warnings

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from bw2data import config, Updates


class UpdaterInterface(object):
    def needed(self):
        return Updates.check_status(False)

    def list(self):
        updates_needed = self.needed()

        if not updates_needed:
            print("\n*** Brightway2 is up to date! ***\n")
        else:
            print("\n*** Updates found ***")
            for update in updates_needed:
                print(Updates.explain(update))
            print("\n*** Action needed ***\nPlease run bw2-uptodate\n")

    def update(self, confirm=True):
        updates_needed = self.needed()

        if updates_needed:
            print("\nThe following updates will be applied:\n")
            for update in updates_needed:
                print(Updates.explain(update))
            if confirm:
                confirmation = input("\nType 'y'to confirm, anything else to cancel: ")
                if confirmation.strip() != "y":
                    print("\n*** Upgrade canceled ***\n")
                    sys.exit(0)

            for update in updates_needed:
                Updates.do_update(update)
        else:
            print("\n*** Brightway2 is up to date! ***\n")


def main():
    args = docopt(__doc__, version="Brightway2 up to date 0.1")
    updater_interface = UpdaterInterface()
    if args["--list"] or args["-l"]:
        updater_interface.list()
    else:
        updater_interface.update()


if __name__ == "__main__":
    main()



from . import config, databases, methods, geomapping
from .parameters import parameters
from .project import projects
import atexit
import random
import shutil
import string
import unittest
import wrapt


class BW2DataTest(unittest.TestCase):
    _multiprocess_can_split_ = True

    def setUp(self):
        config.dont_warn = True
        config.is_test = True
        config.cache = {}
        projects._use_temp_directory()
        self.extra_setup()

    def extra_setup(self):
        pass

    def test_setup_clean(self):
        self.assertEqual(list(databases), [])
        self.assertEqual(list(methods), [])
        self.assertEqual(len(geomapping), 1)  # GLO
        self.assertTrue("GLO" in geomapping)
        self.assertEqual(len(projects), 1)  # Default project
        self.assertTrue("default" in projects)
        self.assertFalse(len(parameters))


@wrapt.decorator
def bw2test(wrapped, instance, args, kwargs):
    config.dont_warn = True
    config.is_test = True
    config.cache = {}
    tempdir = projects._use_temp_directory()
    projects.set_current("".join(random.choices(string.ascii_lowercase, k=18)))
    atexit.register(shutil.rmtree, tempdir)
    return wrapped(*args, **kwargs)



import os


biosphere = {
    ("biosphere", "1"): {
        "categories": ["things"],
        "code": 1,
        "exchanges": [],
        "name": "an emission",
        "type": "emission",
        "unit": "kg",
    },
    ("biosphere", "2"): {
        "categories": ["things"],
        "code": 2,
        "exchanges": [],
        "type": "emission",
        "name": "another emission",
        "unit": "kg",
    },
}

food = {
    ("food", "1"): {
        "categories": ["stuff", "meals"],
        "code": 1,
        "exchanges": [
            {
                "amount": 0.5,
                "input": ("food", "2"),
                "type": "technosphere",
                "uncertainty type": 0,
            },
            {
                "amount": 0.05,
                "input": ("biosphere", "1"),
                "type": "biosphere",
                "uncertainty type": 0,
            },
        ],
        "location": "CA",
        "name": "lunch",
        "type": "process",
        "unit": "kg",
    },
    ("food", "2"): {
        "categories": ["stuff", "meals"],
        "code": 2,
        "exchanges": [
            {
                "amount": 0.25,
                "input": ("food", "1"),
                "type": "technosphere",
                "uncertainty type": 0,
            },
            {
                "amount": 0.15,
                "input": ("biosphere", "2"),
                "type": "biosphere",
                "uncertainty type": 0,
            },
        ],
        "location": "CH",
        "name": "dinner",
        "type": "process",
        "unit": "kg",
    },
}

food2 = {
    ("food", "1"): {
        "categories": ["stuff", "meals"],
        "code": 1,
        "exchanges": [
            {
                "amount": 0.5,
                "input": ("food", "2"),
                "type": "technosphere",
                "uncertainty type": 0,
            },
            {
                "amount": 0.05,
                "input": ("biosphere", "1"),
                "type": "biosphere",
                "uncertainty type": 0,
            },
        ],
        "location": "CA",
        "name": "lunch",
        "type": "process",
        "unit": "kg",
    },
    ("food", "2"): {
        "categories": ["stuff", "meals"],
        "code": 2,
        "exchanges": [
            {
                "amount": 0.25,
                "input": ("food", "1"),
                "type": "technosphere",
                "uncertainty type": 0,
            },
            {
                "amount": 0.15,
                "input": ("biosphere", "2"),
                "type": "biosphere",
                "uncertainty type": 0,
            },
        ],
        "location": "CH",
        "name": "dinner",
        "type": "process",
        "unit": "kg",
    },
}

get_naughty = lambda: [
    x.replace("\n", "")
    for x in open(
        os.path.join(os.path.dirname(__file__), "naughty_strings.txt"), encoding="utf8"
    )
    if x[0] != "#"
]



from .fixtures import biosphere
from bw2data import Database, Method, methods
from bw2data.backends import Activity as PWActivity
from bw2data.errors import ValidityError
from bw2data.tests import BW2DataTest, bw2test
from bw2data.utils import (
    as_uncertainty_dict,
    combine_methods,
    get_activity,
    merge_databases,
    natural_sort,
    random_string,
    uncertainify,
)

# from bw_processing import load_package
import pytest
import stats_arrays as sa


class UtilsTest(BW2DataTest):
    def test_natural_sort(self):
        data = ["s100", "s2", "s1"]
        self.assertEqual(["s1", "s2", "s100"], natural_sort(data))

    def test_random_string(self):
        s = random_string(10)
        self.assertEqual(len(s), 10)
        self.assertTrue(isinstance(s, str))

    def test_combine_methods(self):
        d = Database("biosphere")
        d.register(depends=[])
        d.write(biosphere)
        m1 = Method(("test method 1",))
        m1.register(unit="p")
        m1.write([(("biosphere", 1), 1, "GLO"), (("biosphere", 2), 2, "GLO")])
        m2 = Method(("test method 2",))
        m2.register(unit="p")
        m2.write([(("biosphere", 2), 10, "GLO")])
        combine_methods(("test method 3",), ("test method 1",), ("test method 2",))
        cm = Method(("test method 3",))
        self.assertEqual(
            sorted(cm.load()),
            [(("biosphere", 1), 1, "GLO"), (("biosphere", 2), 12, "GLO")],
        )
        self.assertEqual(methods[["test method 3"]]["unit"], "p")


class UncertainifyTestCase(BW2DataTest):
    def test_wrong_distribution(self):
        with self.assertRaises(AssertionError):
            uncertainify({}, sa.LognormalUncertainty)

    def test_factors_valid(self):
        with self.assertRaises(AssertionError):
            uncertainify({}, bounds_factor=-1)
        with self.assertRaises(TypeError):
            uncertainify({}, bounds_factor="foo")
        with self.assertRaises(AssertionError):
            uncertainify({}, sd_factor=-1)
        with self.assertRaises(TypeError):
            uncertainify({}, sd_factor="foo")

    def test_bounds_factor_none_ok(self):
        uncertainify({}, bounds_factor=None)

    def test_skips(self):
        data = {
            1: {
                "exchanges": [
                    {"type": "production"},
                    {"uncertainty type": sa.LognormalUncertainty.id},
                ]
            }
        }
        # Doesn't raise KeyError for 'amount'
        data = uncertainify(data)

    def test_uniform(self):
        data = {1: {"exchanges": [{"amount": 10.0}]}}
        data = uncertainify(data)
        new_dict = {
            "amount": 10.0,
            "minimum": 9.0,
            "maximum": 11.0,
            "uncertainty type": sa.UniformUncertainty.id,
        }
        self.assertEqual(data[1]["exchanges"][0], new_dict)

    def test_normal_bounded(self):
        data = {1: {"exchanges": [{"amount": 10.0}]}}
        data = uncertainify(data, sa.NormalUncertainty)
        new_dict = {
            "amount": 10.0,
            "loc": 10.0,
            "scale": 1.0,
            "minimum": 9.0,
            "maximum": 11.0,
            "uncertainty type": sa.NormalUncertainty.id,
        }
        self.assertEqual(data[1]["exchanges"][0], new_dict)

    def test_normal_unbounded(self):
        data = {1: {"exchanges": [{"amount": 10.0}]}}
        data = uncertainify(data, sa.NormalUncertainty, bounds_factor=None)
        new_dict = {
            "amount": 10.0,
            "loc": 10.0,
            "scale": 1.0,
            "uncertainty type": sa.NormalUncertainty.id,
        }
        self.assertEqual(data[1]["exchanges"][0], new_dict)

    def test_normal_negative_amount(self):
        data = {1: {"exchanges": [{"amount": -10.0}]}}
        data = uncertainify(data, sa.NormalUncertainty)
        new_dict = {
            "amount": -10.0,
            "loc": -10.0,
            "scale": 1.0,
            "minimum": -11.0,
            "maximum": -9.0,
            "uncertainty type": sa.NormalUncertainty.id,
        }
        self.assertEqual(data[1]["exchanges"][0], new_dict)

    def test_bounds_flipped_negative_amount(self):
        data = {1: {"exchanges": [{"amount": -10.0}]}}
        data = uncertainify(data)
        new_dict = {
            "amount": -10.0,
            "minimum": -11.0,
            "maximum": -9.0,
            "uncertainty type": sa.UniformUncertainty.id,
        }
        self.assertEqual(data[1]["exchanges"][0], new_dict)

    def test_skip_zero_amounts(self):
        data = {1: {"exchanges": [{"amount": 0.0}]}}
        data = uncertainify(data)
        new_dict = {
            "amount": 0.0,
        }
        self.assertEqual(data[1]["exchanges"][0], new_dict)

    def test_get_activity_peewee(self):
        database = Database("a database", "sqlite")
        database.write(
            {
                ("a database", "foo"): {
                    "exchanges": [
                        {
                            "input": ("a database", "foo"),
                            "amount": 1,
                            "type": "production",
                        }
                    ],
                    "location": "bar",
                    "name": "baz",
                },
            }
        )
        self.assertTrue(isinstance(get_activity(("a database", "foo")), PWActivity))


def test_as_uncertainty_dict():
    assert as_uncertainty_dict({}) == {}
    assert as_uncertainty_dict(1) == {"amount": 1.0}
    with pytest.raises(TypeError):
        as_uncertainty_dict("foo")


@bw2test
def test_merge_databases_nonunique_activity_codes():
    first = Database("a database")
    first.write(
        {
            ("a database", "foo"): {
                "exchanges": [
                    {"input": ("a database", "foo"), "amount": 1, "type": "production",}
                ],
                "location": "bar",
                "name": "baz",
            },
        }
    )
    second = Database("another database")
    second.write(
        {
            ("another database", "foo"): {
                "exchanges": [
                    {
                        "input": ("another database", "foo"),
                        "amount": 1,
                        "type": "production",
                    }
                ],
                "location": "bar",
                "name": "baz",
            },
        }
    )
    with pytest.raises(ValidityError):
        merge_databases("a database", "another database")


@bw2test
def test_merge_databases_wrong_backend():
    first = Database("a database", "iotable")
    first.write(
        {
            ("a database", "foo"): {
                "exchanges": [
                    {"input": ("a database", "foo"), "amount": 1, "type": "production",}
                ],
                "location": "bar",
                "name": "baz",
            },
        },
        [],
        [],
        [],
    )
    second = Database("another database")
    second.write(
        {
            ("another database", "bar"): {
                "exchanges": [
                    {
                        "input": ("another database", "bar"),
                        "amount": 1,
                        "type": "production",
                    }
                ],
                "location": "bar",
                "name": "baz",
            },
        }
    )
    with pytest.raises(ValidityError):
        merge_databases("a database", "another database")
    with pytest.raises(ValidityError):
        merge_databases("another database", "a database")


@bw2test
def test_merge_databases_nonexistent():
    first = Database("a database")
    first.write(
        {
            ("a database", "foo"): {
                "exchanges": [
                    {"input": ("a database", "foo"), "amount": 1, "type": "production",}
                ],
                "location": "bar",
                "name": "baz",
            },
        }
    )
    with pytest.raises(AssertionError):
        merge_databases("a database", "another database")
    with pytest.raises(AssertionError):
        merge_databases("another database", "a database")


# @bw2test
# def test_merge_databases():
#     first = Database("a database")
#     first.write(
#         {
#             ("a database", "foo"): {
#                 "exchanges": [
#                     {"input": ("a database", "foo"), "amount": 1, "type": "production",}
#                 ],
#                 "location": "bar",
#                 "name": "baz",
#             },
#         }
#     )
#     second = Database("another database")
#     second.write(
#         {
#             ("another database", "bar"): {
#                 "exchanges": [
#                     {
#                         "input": ("another database", "bar"),
#                         "amount": 1,
#                         "type": "production",
#                     }
#                 ],
#                 "location": "bar",
#                 "name": "baz",
#             },
#         }
#     )
#     merge_databases("a database", "another database")
#     merged = Database("a database")
#     assert len(merged) == 2
#     assert "another database" not in databases
#     assert ("a database", "bar") in mapping
#     print(merged.filepath_processed())
#     package = load_package(merged.filepath_processed())
#     print(package.keys())
#     array = package["technosphere_matrix.npy"]
#     assert mapping[("a database", "bar")] in {x["col_value"] for x in array}
#     assert mapping[("a database", "foo")] in {x["col_value"] for x in array}



# The MIT License (MIT)

# Copyright (c) 2014 abarnert

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# fatomic from abarnert
# https://github.com/abarnert/fatomic
import contextlib
import errno
import os
import shutil
import sys
import tempfile
import types

__all__ = [
    "replace",
    "open",
    "writeall",
    "writechunks",
    "writelines",
    "transformall",
    "transformchunks",
    "transformlines",
    "transform",
]

# If we're on 3.3+, just use os.replace; if we're on POSIX, rename
# and replace do the same thing.
try:
    replace = os.replace
except AttributeError:
    if sys.platform != "win32":
        replace = os.rename
    else:
        # This requires PyWin32 if you're on Windows. If that's not
        # accepted, you can write a ctypes solution, but then you'll
        # have to handle unicode-vs.-bytes strings and creating an
        # OSError from GetLastError and so on yourself, which I don't
        # feel like doing. (I'll accept a pull request from anyone
        # else who does...)
        import win32api, win32con

        def replace(src, dst):
            win32api.MoveFileEx(src, dst, win32con.MOVEFILE_REPLACE_EXISTING)


def _guessmode(contents, binary):
    if binary is not None:
        return binary
    if isinstance(contents, str):
        return False
    elif isinstance(contents, (bytes, bytearray)):
        return True
    else:
        return False


def _mode(binary, contents=None):
    if binary is not None:
        return "b" if binary else ""
    if contents is None:
        raise TypeError("binary cannot be None")
    if isinstance(contents, str):
        return ""
    elif isinstance(contents, (bytes, bytearray)):
        return "b"
    else:
        return ""


def _tempfile(filename, mode):
    return tempfile.NamedTemporaryFile(
        mode=mode, prefix=os.path.basename(filename), suffix=".tmp", delete=False
    )


_open = open


@contextlib.contextmanager
def open(filename, mode, *args, **kwargs):
    if mode[0] not in "wxa" or len(mode) > 1 and mode[1] == "+":
        raise ValueError("invalid mode: '{}'".format(mode))
    f = _tempfile(filename, mode, *args, **kwargs)
    _discard = [False]
    try:
        if mode[0] == "a":
            try:
                with _open(filename, "r" + mode[1:], *args, **kwargs) as fin:
                    shutil.copyfileobj(fin, f)
            except (OSError, IOError) as e:
                if e.errno == errno.ENOENT:
                    pass

        def discard(self, _discard=_discard):
            _discard[0] = True

        f.discard = types.MethodType(discard, f)
        yield f
    finally:
        f.close()
        if not _discard[0]:
            try:
                replace(f.name, filename)
            except OSError:
                shutil.move(f.name, filename)


def write(filename, lines, binary=False):
    mode = _mode(binary)
    with open(filename, "w" + mode) as fout:
        fout.writelines(lines)


def writeall(filename, contents, binary=None):
    mode = _mode(binary, contents)
    with open(filename, "w" + mode) as fout:
        f.write(contents)


writechunks = write


def transform(filename, func, binary=False):
    mode = _mode(binary)
    with open(filename, "w" + mode) as fout, _open(filename, "r" + mode) as fin:
        fout.writelines(func(line) for line in fin)


def transformall(filename, func, binary=False):
    mode = _mode(binary)
    with open(filename, "w" + mode) as fout, _open(filename, "r" + mode) as fin:
        fout.writelines(fin.read())


def _chunkfile(f, chunksize=None):
    if chunksize is None:
        chunksize = 8192
    while True:
        buf = f.read(chunksize)
        if not buf:
            return
        yield buf


def transformchunks(filename, func, chunksize=None, binary=False):
    mode = _mode(binary)
    with open(filename, "w" + mode) as fout, _open(filename, "r" + mode) as fin:
        fout.writelines(_chunkfile(fin, chunksize))


def append(filename, lines, binary=False):
    mode = _mode(binary)
    with open(filename, "a" + mode) as fout:
        fout.writelines(lines)


def appendall(filename, contents, binary=False):
    mode = _mode(binary)
    with open(filename, "a" + mode) as fout:
        fout.write(contents)


appendchunks = append


def test():
    with open("foo", "a") as f:
        f.write("Hello, ")
    with open("foo", "a") as f:
        f.write("world\n")
    transform("foo", lambda line: line.replace("Hello", "Hi"))


if __name__ == "__main__":
    test()



from bw2data.tests import bw2test
from bw2data import (
    geomapping,
    databases,
    methods,
    Method,
    projects,
    get_activity,
)
from bw2data.database import DatabaseChooser
from bw2data.parameters import (
    ActivityParameter,
    ParameterizedExchange,
    parameters,
)

try:
    import bw2calc
except ImportError:
    bw2calc = None
import numpy as np
import stats_arrays as sa
import pytest


@pytest.fixture
@bw2test
def activity():
    database = DatabaseChooser("db")
    database.write(
        {
            ("db", "a"): {
                "exchanges": [
                    {"input": ("db", "a"), "amount": 2, "type": "production",},
                    {"input": ("db", "b"), "amount": 3, "type": "technosphere",},
                    {"input": ("db", "c"), "amount": 4, "type": "biosphere",},
                ],
                "name": "a",
            },
            ("db", "b"): {"name": "b"},
            ("db", "c"): {"name": "c", "type": "biosphere"},
            ("db", "d"): {
                "name": "d",
                "exchanges": [
                    {"input": ("db", "a"), "amount": 5, "type": "technosphere"},
                    {"input": ("db", "b"), "amount": -0.1, "type": "substitution"},
                ],
            },
        }
    )
    return database.get("a")


@pytest.fixture
@bw2test
def activity_and_method():
    database = DatabaseChooser("db")
    database.write(
        {
            ("db", "a"): {
                "exchanges": [
                    {"input": ("db", "a"), "amount": 2, "type": "production",},
                    {"input": ("db", "b"), "amount": 3, "type": "technosphere",},
                    {"input": ("db", "c"), "amount": 4, "type": "biosphere",},
                ],
                "name": "a",
            },
            ("db", "b"): {"name": "b"},
            ("db", "c"): {"name": "c", "type": "biosphere"},
            ("db", "d"): {
                "name": "d",
                "exchanges": [
                    {"input": ("db", "a"), "amount": 5, "type": "technosphere"}
                ],
            },
        }
    )
    cfs = [(("db", "c"), 42)]
    method = Method(("a method",))
    method.register()
    method.write(cfs)
    return database.get("a"), method


def test_setup_clean(activity):
    assert len(databases) == 1
    assert list(methods) == []
    assert len(geomapping) == 1  # GLO
    assert "GLO" in geomapping
    assert len(projects) == 2  # Default project
    assert "default" in projects


def test_production(activity):
    assert len(list(activity.production())) == 1
    assert len(activity.production()) == 1
    exc = list(activity.production())[0]
    assert exc["amount"] == 2


def test_substitution(activity):
    d = get_activity(("db", "d"))
    assert len(activity.substitution()) == 0
    assert len(d.substitution()) == 1


def test_biosphere(activity):
    assert len(list(activity.biosphere())) == 1
    assert len(activity.biosphere()) == 1
    exc = list(activity.biosphere())[0]
    assert exc["amount"] == 4


def test_technosphere(activity):
    assert len(list(activity.technosphere())) == 1
    assert len(activity.technosphere()) == 1
    exc = list(activity.technosphere())[0]
    assert exc["amount"] == 3


def test_technosphere_include_substitution(activity):
    d = get_activity(("db", "d"))
    assert len(d.technosphere()) == 2
    assert len(d.technosphere(include_substitution=True)) == 2


def test_technosphere_exclude_substitution(activity):
    d = get_activity(("db", "d"))
    assert len(d.technosphere(include_substitution=False)) == 1


def test_upstream(activity):
    assert len(list(activity.upstream())) == 1
    assert len(activity.upstream()) == 1
    exc = list(activity.upstream())[0]
    assert exc["amount"] == 5


def test_upstream_no_kinds(activity):
    act = get_activity(("db", "c"))
    assert len(list(act.upstream(kinds=None))) == 1
    assert len(act.upstream(kinds=None)) == 1
    exc = list(act.upstream(kinds=None))[0]
    assert exc["amount"] == 4


def test_upstream_bio(activity):
    act = get_activity(("db", "c"))
    assert len(list(act.upstream())) == 0
    assert len(act.upstream()) == 0


def test_ordering_consistency(activity):
    ordering = [[exc["amount"] for exc in activity.exchanges()] for _ in range(100)]
    for sample in ordering[1:]:
        assert sample == ordering[0]


@bw2test
def test_uncertainty():
    database = DatabaseChooser("db")
    database.write(
        {
            ("db", "a"): {
                "exchanges": [
                    {
                        "input": ("db", "a"),
                        "amount": 2,
                        "type": "production",
                        "uncertainty type": 1,
                        "loc": 2,
                        "scale": 3,
                        "shape": 4,
                        "maximum": 5,
                        "negative": True,
                        "name": "a",
                    }
                ]
            }
        }
    )
    expected = {
        "uncertainty type": 1,
        "loc": 2,
        "scale": 3,
        "shape": 4,
        "maximum": 5,
        "negative": True,
    }
    exchange = list(database.get("a").exchanges())[0]
    assert exchange.uncertainty == expected


@bw2test
def test_uncertainty_type():
    database = DatabaseChooser("db")
    database.write(
        {
            ("db", "a"): {
                "exchanges": [
                    {
                        "input": ("db", "a"),
                        "amount": 2,
                        "type": "production",
                        "uncertainty type": 1,
                        "name": "a",
                    }
                ]
            }
        }
    )
    exchange = list(database.get("a").exchanges())[0]
    assert exchange.uncertainty_type.id == 1
    assert exchange.uncertainty_type == sa.NoUncertainty


@bw2test
def test_uncertainty_type_missing():
    database = DatabaseChooser("db")
    database.write(
        {
            ("db", "a"): {
                "exchanges": [
                    {
                        "input": ("db", "a"),
                        "amount": 2,
                        "type": "production",
                        "name": "a",
                    }
                ]
            }
        }
    )
    exchange = list(database.get("a").exchanges())[0]
    assert exchange.uncertainty_type.id == 0
    assert exchange.uncertainty_type == sa.UndefinedUncertainty


@bw2test
def test_random_sample():
    database = DatabaseChooser("db")
    database.write(
        {
            ("db", "a"): {
                "exchanges": [
                    {
                        "input": ("db", "a"),
                        "type": "production",
                        "amount": 20,
                        "unit": "kg",
                        "type": "biosphere",
                        "uncertainty type": 2,
                        "loc": np.log(20),
                        "scale": 1.01,
                    }
                ],
                "name": "a",
            }
        }
    )
    exchange = list(database.get("a").exchanges())[0]
    assert (exchange.random_sample() > 0).sum() == 100
    assert exchange.random_sample().shape == (100,)


@bw2test
def test_random_sample_negative():
    database = DatabaseChooser("db")
    database.write(
        {
            ("db", "a"): {
                "exchanges": [
                    {
                        "input": ("db", "a"),
                        "type": "production",
                        "amount": -20,
                        "negative": True,
                        "unit": "kg",
                        "type": "biosphere",
                        "uncertainty type": 2,
                        "loc": np.log(20),
                        "scale": 1.01,
                    }
                ],
                "name": "a",
            }
        }
    )
    exchange = list(database.get("a").exchanges())[0]
    assert (exchange.random_sample() < 0).sum() == 100
    assert exchange.random_sample().shape == (100,)


@pytest.mark.skip()
@pytest.mark.skipif(bw2calc is None, reason="requires bw2calc")
def test_lca(activity_and_method):
    a, m = activity_and_method
    exc = list(a.production())[0]
    lca = exc.lca(method=m.name)
    assert lca.score == 4 * 42
    lca = exc.lca(method=m.name, amount=1)
    assert lca.score == 2 * 42


@bw2test
def test_delete_parameterized_exchange():
    db = DatabaseChooser("example")
    db.register()

    a = db.new_activity(code="A", name="An activity")
    a.save()
    b = db.new_activity(code="B", name="Another activity")
    b.save()
    exc = a.new_exchange(
        amount=0, input=b, type="technosphere", formula="foo * bar + 4"
    )
    exc.save()

    activity_data = [
        {
            "name": "reference_me",
            "formula": "sqrt(25)",
            "database": "example",
            "code": "B",
        },
        {
            "name": "bar",
            "formula": "reference_me + 2",
            "database": "example",
            "code": "A",
        },
    ]
    parameters.new_activity_parameters(activity_data, "my group")
    parameters.add_exchanges_to_group("my group", a)

    assert ActivityParameter.select().count() == 2
    assert ParameterizedExchange.select().count() == 1

    exc.delete()
    assert ActivityParameter.select().count() == 2
    assert not ParameterizedExchange.select().count()



from . import (
    Database,
    databases,
    normalizations,
    Normalization,
    weightings,
    Weighting,
    methods,
    Method,
    projects
)
from .errors import Brightway2Project
from .backends.schema import ActivityDataset as AD, get_id
from fs.zipfs import ZipFS


class Mapping:
    """A dictionary that maps object ids, like ``("Ecoinvent 2.2", 42)``, to integers.

    Used only for backwards compatibility; preferred method is now to look up the ids of activities directly in the SQlite database."""

    def add(self, keys):
        raise DeprecationWarning(
            "This method is no longer necessary, and does nothing."
        )
        return

    def __getitem__(self, key):
        return get_id(key)

    def delete(self, keys):
        raise DeprecationWarning(
            "This method is no longer necessary, and does nothing."
        )
        return

    def __str__(self):
        return "Obsolete mapping dictionary."

    def __len__(self):
        return AD.select().count()


def unpack(dct):
    for obj in dct:
        if hasattr(obj, "key"):
            yield obj.key
        else:
            yield obj


def translate_key(key):
    if isinstance(key, int):
        return key
    else:
        return AD.get(AD.database == key[0], AD.code == key[1]).id


def prepare_lca_inputs(
    demand=None,
    method=None,
    weighting=None,
    normalization=None,
    demands=None,
    remapping=True,
    demand_database_last=True,
):
    """Prepare LCA input arguments in Brightway 2.5 style."""
    if not projects.dataset.data.get("25"):
        raise Brightway2Project("Please use `projects.migrate_project_25` before calculating using Brightway 2.5")

    databases.clean()
    data_objs = []

    if demands:
        demand_database_names = [db_label for dct in demands for db_label, _ in unpack(dct)]
    elif demand:
        demand_database_names = [db_label for db_label, _ in unpack(demand)]
    else:
        raise ValueError("Need some form of demand for LCA calculation")

    database_names = set.union(
        *[
            Database(db_label).find_graph_dependents()
            for db_label in demand_database_names
        ]
    )

    if demand_database_last:
        database_names = [x for x in database_names if x not in demand_database_names] + demand_database_names

    data_objs.extend(
        [ZipFS(Database(obj).filepath_processed()) for obj in database_names]
    )

    if method:
        assert method in methods
        data_objs.append(ZipFS(Method(method).filepath_processed()))
    if weighting:
        assert weighting in weightings
        data_objs.append(ZipFS(Weighting(weighting).filepath_processed()))
    if normalization:
        assert normalization in normalizations
        data_objs.append(ZipFS(Normalization(normalization).filepath_processed()))

    if remapping:
        reversed_mapping = {
            i: (d, c)
            for d, c, i in AD.select(AD.database, AD.code, AD.id)
            .where(AD.database << database_names)
            .tuples()
        }
        remapping_dicts = {
            "activity": reversed_mapping,
            "product": reversed_mapping,
            "biosphere": reversed_mapping,
        }
    else:
        remapping_dicts = {}

    if demands:
        indexed_demand = [{get_id(k): v for k, v in dct.items()} for dct in demands]
    elif demand:
        indexed_demand = {get_id(k): v for k, v in demand.items()}
    else:
        indexed_demand = None

    return indexed_demand, data_objs, remapping_dicts


def get_database_filepath(functional_unit):
    """Get filepaths for all databases in supply chain of `functional_unit`"""
    dbs = set.union(
        *[Database(key[0]).find_graph_dependents() for key in functional_unit]
    )
    return [Database(obj).filepath_processed() for obj in dbs]



from bw2data.tests import bw2test
from bw2data import (
    config,
    databases,
    geomapping,
    mapping,
    methods,
    preferences,
    projects,
)
from bw2data.project import ProjectDataset
from bw2data.errors import ReadOnlyProject
from pathlib import Path
import os
import pytest
import tempfile


###
### Basic setup
###


@bw2test
def test_project_directories():
    projects.set_current("foo")
    for dirname in projects._basic_directories:
        assert os.path.isdir(os.path.join(projects.dir, dirname))


@bw2test
def test_from_env_var():
    dirpath = tempfile.mkdtemp()
    os.environ["BRIGHTWAY2_DIR"] = dirpath
    base, logs = projects._get_base_directories()
    assert os.path.isdir(base)
    assert os.path.isdir(logs)
    del os.environ["BRIGHTWAY2_DIR"]


@bw2test
def test_invalid_env_var():
    os.environ["BRIGHTWAY2_DIR"] = "nothing special"
    with pytest.raises(OSError):
        projects._get_base_directories()
    del os.environ["BRIGHTWAY2_DIR"]


@bw2test
def test_invalid_output_env_dir():
    os.environ["BRIGHTWAY2_OUTPUT_DIR"] = "nothing special"
    assert str(projects.dir) in str(projects.output_dir)
    del os.environ["BRIGHTWAY2_OUTPUT_DIR"]


@bw2test
def test_output_env_dir():
    assert Path(os.getcwd()) != projects.output_dir
    os.environ["BRIGHTWAY2_OUTPUT_DIR"] = os.getcwd()
    assert Path(os.getcwd()) == projects.output_dir
    del os.environ["BRIGHTWAY2_OUTPUT_DIR"]


@bw2test
def test_output_dir_from_preferences():
    assert Path(os.getcwd()) != projects.output_dir
    preferences["output_dir"] = os.getcwd()
    assert Path(preferences["output_dir"]) == projects.output_dir


@bw2test
def test_invalid_output_dir_from_preferences():
    preferences["output_dir"] = "nope"
    assert str(projects.dir) in str(projects.output_dir)
    del preferences["output_dir"]


@bw2test
def test_directories():
    assert os.path.isdir(projects.dir)
    assert os.path.isdir(projects.logs_dir)


@bw2test
def test_default_project_created():
    assert "default" in projects


@bw2test
def test_repeatedly_set_name_same_value():
    projects.set_current("foo")
    projects.set_current("foo")
    projects.set_current("foo")
    assert len(projects) == 3
    assert "foo" in projects


@pytest.mark.skipif(config._windows, reason="Windows doesn't allow fun")
@bw2test
def test_funny_project_names():
    NAMES = [
        "Powerلُلُصّبُلُلصّبُررً ॣ ॣh ॣ ॣ冗",
        "Roses are [0;31mred[0m, violets are [0;34mblue. Hope you enjoy terminal hue",
        "True",
        "None",
        "1.0/0.0",
        "0xabad1dea",
        "!@#$%^&*()`~",
        "<>?:'{}|_+",
        r",./;'[]\-=",
        "Ω≈ç√∫˜µ≤≥÷",
        "田中さんにあげて下さい",
        "｀ｨ(´∀｀∩",
        "👾 🙇 💁 🙅 🙆 🙋 🙎 🙍 ",
        "הָיְתָהtestالصفحات التّحول",
        "　",
    ]
    error_found = False
    for name in NAMES:
        try:
            projects.set_current(name)
            assert [x for x in os.listdir(projects.dir)]
            print("This is OK:", name)
        except:
            print("This is not OK:", name)
            error_found = True
    if error_found:
        raise ValueError("Oops")


@bw2test
def test_report():
    assert projects.report


@bw2test
def test_request_directory():
    projects.request_directory("foo")
    assert "foo" in os.listdir(projects.dir)


###
### Project deletion
###


@bw2test
def test_delete_current_project_with_name():
    projects.set_current("foo")
    projects.delete_project("foo")
    assert projects.current != "foo"
    assert "foo" not in projects


@bw2test
def test_delete_project_remove_directory():
    projects.set_current("foo")
    foo_dir = projects.dir
    projects.set_current("bar")
    projects.delete_project("foo", delete_dir=True)
    assert not os.path.isdir(foo_dir)
    assert "foo" not in projects
    assert projects.current == "bar"


@bw2test
def test_delete_project_keep_directory():
    projects.set_current("foo")
    foo_dir = projects.dir
    projects.set_current("bar")
    projects.delete_project("foo")
    assert os.path.isdir(foo_dir)
    assert "foo" not in projects
    assert projects.current == "bar"


@bw2test
def test_delete_project():
    projects.set_current("foo")
    projects.set_current("bar")
    projects.delete_project("foo")
    assert "foo" not in projects
    assert projects.current == "bar"


@bw2test
def test_delete_last_project():
    assert len(projects) == 2
    with pytest.raises(ValueError):
        projects.delete_project()
        projects.delete_project()


@bw2test
def test_delete_current_project_no_name():
    projects.set_current("foo")
    projects.delete_project()
    assert "foo" not in projects
    assert projects.current != "foo"


###
### Set project
###


@bw2test
def test_error_outdated_set_project():
    assert projects.current
    with pytest.raises(AttributeError):
        projects.current = "Foo"


@bw2test
def test_set_project_creates_new_project():
    other_num = len(projects)
    projects.set_current("foo")
    assert len(projects) == other_num + 1


@bw2test
def test_set_project():
    projects.set_current("foo")
    assert projects.current == "foo"


@bw2test
def test_set_project_default_writable():
    pass


@bw2test
def test_set_project_writable_even_if_writable_false():
    pass


@bw2test
def test_set_readonly_project():
    projects.set_current("foo")
    assert not projects.read_only
    config.p["lockable"] = True
    projects.set_current("foo", writable=False)
    assert projects.read_only


@bw2test
def test_set_readonly_project_first_time():
    projects.set_current("foo", writable=False)
    assert not projects.read_only


@bw2test
def test_set_current_reset_metadata():
    databases["foo"] = "bar"
    assert "foo" in databases
    projects.set_current("foo")
    assert "foo" not in databases


###
### Test magic methods
###


@bw2test
def test_representation():
    assert repr(projects)
    assert str(projects)


@bw2test
def test_contains():
    assert "default" in projects
    projects.set_current("foo")
    assert "foo" in projects


@bw2test
def test_len():
    assert len(projects) == 2
    projects.set_current("foo")
    assert len(projects) == 3


@bw2test
def test_iterating_over_projects_no_error():
    projects.set_current("foo")
    projects.set_current("bar")
    projects.set_current("baz")
    for x in projects:
        projects.set_current(x.name)


###
### Read-only constraints
###


@bw2test
def test_create_lock_file():
    projects.set_current("foo")
    config.p["lockable"] = True
    projects.set_current("foo", writable=False)
    assert not os.path.isfile(os.path.join(projects.dir, "write-lock"))
    projects.set_current("bar")
    assert not os.path.isfile(os.path.join(projects.dir, "write-lock"))
    config.p["lockable"] = True
    projects.set_current("bar")
    assert os.path.isfile(os.path.join(projects.dir, "write-lock"))


@bw2test
def test_lockable_config_missing():
    assert "lockable" not in config.p


@bw2test
def test_read_only_cant_write():
    projects.set_current("foo")
    config.p["lockable"] = True
    projects.set_current("foo", writable=False)
    with pytest.raises(ReadOnlyProject):
        databases["foo"] = "bar"


###
### Copy project
###


@bw2test
def test_copy_project():
    ds = ProjectDataset.get(ProjectDataset.name == projects.current)
    ds.data["this"] = "that"
    ds.save()

    databases["foo"] = "bar"
    projects.copy_project("another one", False)
    assert "another one" in projects

    ds = ProjectDataset.get(ProjectDataset.name == "another one")
    assert ds.data["this"] == "that"

    projects.set_current("another one")
    assert databases["foo"] == "bar"


@bw2test
def test_copy_project_switch_current():
    projects.copy_project("another one")
    assert projects.current == "another one"


# TODO: purge delete directories



from .. import projects
from .schema import bw2_schema
from whoosh import index
import os
import shutil


class IndexManager(object):
    def __init__(self, database_path, dir_name="whoosh"):
        self.path = os.path.join(projects.request_directory("whoosh"), database_path)
        if not os.path.exists(self.path):
            os.mkdir(self.path)

    def get(self):
        try:
            return index.open_dir(self.path)
        except index.EmptyIndexError:
            return self.create()

    def create(self):
        self.delete_database()
        os.mkdir(self.path)
        return index.create_in(self.path, bw2_schema)

    def _format_dataset(self, ds):
        fl = lambda o: o[1].lower() if isinstance(o, tuple) else o.lower()
        return dict(
            name=ds.get("name", "").lower(),
            comment=ds.get("comment", "").lower(),
            product=ds.get("reference product", "").lower(),
            categories=u", ".join(ds.get("categories", [])).lower(),
            synonyms=u", ".join(ds.get("synonyms", [])).lower(),
            location=fl(ds.get("location", "")),
            database=ds["database"],
            code=ds["code"],
        )

    def add_dataset(self, ds):
        self.add_datasets([ds])

    def add_datasets(self, datasets):
        writer = self.get().writer()
        for ds in datasets:
            writer.add_document(**self._format_dataset(ds))
        writer.commit()

    def update_dataset(self, ds):
        writer = self.get().writer()
        writer.update_document(**self._format_dataset(ds))
        writer.commit()

    def delete_dataset(self, ds):
        index = self.get()
        index.delete_by_term("code", ds["code"])

    def delete_database(self):
        shutil.rmtree(self.path)



__all__ = [
    "dynamic_calculation_setups",
    "calculation_setups",
    "config",
    "convert_backend",
    "Database",
    "databases",
    "DataStore",
    "get_activity",
    "get_id",
    "geomapping",
    "IndexManager",
    "JsonWrapper",
    "mapping",
    "Method",
    "methods",
    "Normalization",
    "normalizations",
    "parameters",
    "preferences",
    "prepare_lca_inputs",
    "ProcessedDataStore",
    "projects",
    "Searcher",
    "set_data_dir",
    "Weighting",
    "weightings",
]

from .version import version as __version__

from .configuration import config
from .project import projects
from .utils import set_data_dir
from .meta import (
    dynamic_calculation_setups,
    calculation_setups,
    databases,
    geomapping,
    methods,
    normalizations,
    preferences,
    weightings,
)

# Add metadata class instances to global list of serialized metadata
config.metadata.extend(
    [
        dynamic_calculation_setups,
        calculation_setups,
        databases,
        geomapping,
        methods,
        normalizations,
        preferences,
        weightings,
    ]
)

# Backwards compatibility - preferable to access ``preferences`` directly
config.p = preferences

from .serialization import JsonWrapper
from .database import DatabaseChooser as Database
from .utils import get_activity
from .data_store import DataStore, ProcessedDataStore
from .method import Method
from .search import Searcher, IndexManager
from .weighting_normalization import Weighting, Normalization
from .backends import convert_backend, get_id
from .compat import prepare_lca_inputs, Mapping

mapping = Mapping()

from .updates import Updates
from .parameters import parameters

Updates.check_status()



from bw2data.serialization import JsonSanitizer
import unittest


class JsonSantizierTestCase(unittest.TestCase):
    def test_tuple(self):
        self.assertEqual(JsonSanitizer.sanitize((1,)), {"__tuple__": True, "data": [1]})
        self.assertEqual(JsonSanitizer.load({"__tuple__": True, "data": [1]}), (1,))

    def test_dict(self):
        self.assertEqual(
            JsonSanitizer.sanitize({1: 2}),
            {"__dict__": True, "keys": [1], "values": [2]},
        )
        self.assertEqual(
            JsonSanitizer.load({"__dict__": True, "keys": [1], "values": [2]}), {1: 2}
        )

    def test_nested(self):
        input_data = [{(1, 2): "foo"}, ["bar", (5, 6)], {}, tuple([]), ((7,),)]
        expected = [
            {
                "__dict__": True,
                "keys": [{"__tuple__": True, "data": [1, 2]}],
                "values": ["foo"],
            },
            ["bar", {"__tuple__": True, "data": [5, 6]}],
            {"__dict__": True, "keys": [], "values": []},
            {"__tuple__": True, "data": []},
            {"__tuple__": True, "data": [{"__tuple__": True, "data": [7]}]},
        ]
        self.assertEqual(JsonSanitizer.sanitize(input_data), expected)
        self.assertEqual(JsonSanitizer.load(expected), input_data)



import hashlib
import os
import re

re_slugify = re.compile(r"[^\w\s-]", re.UNICODE)


def create_dir(dirpath):
    "Create directory tree to `dirpath`; ignore if already exists"
    if not os.path.isdir(dirpath):
        os.makedirs(dirpath)


def check_dir(directory):
    """Returns ``True`` if given path is a directory and writeable, ``False`` otherwise."""
    return os.path.isdir(directory) and os.access(directory, os.W_OK)


def md5(filepath, blocksize=65536):
    """Generate MD5 hash for file at `filepath`"""
    hasher = hashlib.md5()
    fo = open(filepath, "rb")
    buf = fo.read(blocksize)
    while len(buf) > 0:
        hasher.update(buf)
        buf = fo.read(blocksize)
    return hasher.hexdigest()



from . import sqlite3_lci_db
from .. import databases, geomapping
from ..errors import ValidityError
from ..project import writable_project
from ..proxies import ActivityProxyBase, ExchangeProxyBase
from ..search import IndexManager
from .schema import ActivityDataset, ExchangeDataset
from .utils import dict_as_activitydataset, dict_as_exchangedataset
from collections.abc import Iterable
import copy
import uuid


class Exchanges(Iterable):
    """Iterator for exchanges with some additional methods.

    This is not a generator; ``next()`` is not supported. Everything time you start to iterate over the object you get a new list starting from the beginning. However, to get a single item you can do ``next(iter(foo))``.

    Ordering is by database row id.

    Supports the following:

    .. code-block:: python

        exchanges = activity.exchanges()

        # Iterate
        for exc in exchanges:
            pass

        # Length
        len(exchanges)

        # Delete all
        exchanges.delete()

    """

    def __init__(self, key, kinds=None, reverse=False):
        self._key, self._kinds = key, kinds
        if reverse:
            self._args = [
                ExchangeDataset.input_database == self._key[0],
                ExchangeDataset.input_code == self._key[1],
                # No production exchanges - these two clauses have to be together,
                # not individually
                ExchangeDataset.output_database
                != self._key[0] & ExchangeDataset.output_code
                != self._key[1],
            ]
        else:
            self._args = [
                ExchangeDataset.output_database == self._key[0],
                ExchangeDataset.output_code == self._key[1],
            ]
        if self._kinds:
            self._args.append(ExchangeDataset.type << self._kinds)

    def filter(self, expr):
        self._args.append(expr)

    @writable_project
    def delete(self):
        databases.set_dirty(self._key[0])
        ExchangeDataset.delete().where(*self._args).execute()

    def _get_queryset(self):
        return ExchangeDataset.select().where(*self._args).order_by(ExchangeDataset.id)

    def __iter__(self):
        for obj in self._get_queryset():
            yield Exchange(obj)

    def __len__(self):
        return self._get_queryset().count()


class Activity(ActivityProxyBase):
    def __init__(self, document=None, **kwargs):
        """Create an `Activity` proxy object.

        If this is a new activity, can pass `kwargs`.

        If the activity exists in the database, `document` should be an `ActivityDataset`."""
        if document is None:
            self._document = ActivityDataset()
            self._data = kwargs
        else:
            self._document = document
            self._data = self._document.data
            self._data["code"] = self._document.code
            self._data["database"] = self._document.database
            self._data["id"] = self._document.id

    @property
    def id(self):
        return self._document.id

    def __setitem__(self, key, value):
        if key == "id":
            raise ValueError("`id` is read-only")
        elif key == "code" and "code" in self._data:
            self._change_code(value)
            print(
                "Successfully switched activity dataset to new code `{}`".format(value)
            )
        elif key == "database" and "database" in self._data:
            self._change_database(value)
            print("Successfully switch activity dataset to database `{}`".format(value))
        else:
            super(Activity, self).__setitem__(key, value)

    @property
    def key(self):
        return (self.get("database"), self.get("code"))

    @writable_project
    def delete(self):
        from .. import Database
        from ..parameters import ActivityParameter, ParameterizedExchange

        try:
            ap = ActivityParameter.get(database=self[0], code=self[1])
            ParameterizedExchange.delete().where(
                ParameterizedExchange.group == ap.group
            ).execute()
            ActivityParameter.delete().where(
                ActivityParameter.database == self[0], ActivityParameter.code == self[1]
            ).execute()
        except ActivityParameter.DoesNotExist:
            pass
        IndexManager(Database(self["database"]).filename).delete_dataset(self._data)
        self.exchanges().delete()
        self._document.delete_instance()
        self = None

    @writable_project
    def save(self):
        from .. import Database

        if not self.valid():
            raise ValidityError(
                "This activity can't be saved for the "
                "following reasons\n\t* " + "\n\t* ".join(self.valid(why=True)[1])
            )

        databases.set_dirty(self["database"])

        for key, value in dict_as_activitydataset(self._data).items():
            if key != "id":
                setattr(self._document, key, value)
        self._document.save()

        if self.get("location") and self["location"] not in geomapping:
            geomapping.add([self["location"]])

        if databases[self["database"]].get("searchable", True):
            IndexManager(Database(self["database"]).filename).update_dataset(self._data)

    def _change_code(self, new_code):
        if self["code"] == new_code:
            return

        if (
            ActivityDataset.select()
            .where(
                ActivityDataset.database == self["database"],
                ActivityDataset.code == new_code,
            )
            .count()
        ):
            raise ValueError(
                "Activity database with code `{}` already exists".format(new_code)
            )

        with sqlite3_lci_db.atomic() as txn:
            ActivityDataset.update(code=new_code).where(
                ActivityDataset.database == self["database"],
                ActivityDataset.code == self["code"],
            ).execute()
            ExchangeDataset.update(output_code=new_code).where(
                ExchangeDataset.output_database == self["database"],
                ExchangeDataset.output_code == self["code"],
            ).execute()
            ExchangeDataset.update(input_code=new_code).where(
                ExchangeDataset.input_database == self["database"],
                ExchangeDataset.input_code == self["code"],
            ).execute()

        if databases[self["database"]].get("searchable"):
            from .. import Database

            IndexManager(Database(self["database"]).filename).delete_dataset(self)
            self._data["code"] = new_code
            IndexManager(Database(self["database"]).filename).add_datasets([self])
        else:
            self._data["code"] = new_code

    def _change_database(self, new_database):
        if self["database"] == new_database:
            return

        if new_database not in databases:
            raise ValueError("Database {} does not exist".format(new_database))

        with sqlite3_lci_db.atomic() as txn:
            ActivityDataset.update(database=new_database).where(
                ActivityDataset.database == self["database"],
                ActivityDataset.code == self["code"],
            ).execute()
            ExchangeDataset.update(output_database=new_database).where(
                ExchangeDataset.output_database == self["database"],
                ExchangeDataset.output_code == self["code"],
            ).execute()
            ExchangeDataset.update(input_database=new_database).where(
                ExchangeDataset.input_database == self["database"],
                ExchangeDataset.input_code == self["code"],
            ).execute()

        if databases[self["database"]].get("searchable"):
            from .. import Database

            IndexManager(Database(self["database"]).filename).delete_dataset(self)
            self._data["database"] = new_database
            IndexManager(Database(self["database"]).filename).add_datasets([self])
        else:
            self._data["database"] = new_database

    def exchanges(self):
        return Exchanges(self.key)

    def technosphere(self, include_substitution=True):
        return Exchanges(
            self.key,
            kinds=(
                ("technosphere", "substitution")
                if include_substitution
                else ("technosphere",)
            ),
        )

    def biosphere(self):
        return Exchanges(self.key, kinds=("biosphere",),)

    def production(self):
        return Exchanges(self.key, kinds=("production",),)

    def substitution(self):
        return Exchanges(self.key, kinds=("substitution",),)

    def upstream(self, kinds=("technosphere",)):
        return Exchanges(self.key, kinds=kinds, reverse=True)

    def new_exchange(self, **kwargs):
        """Create a new exchange linked to this activity"""
        exc = Exchange()
        exc.output = self.key
        for key in kwargs:
            exc[key] = kwargs[key]
        return exc

    @writable_project
    def copy(self, code=None, **kwargs):
        """Copy the activity. Returns a new `Activity`.

        `code` is the new activity code; if not given, a UUID is used.

        `kwargs` are additional new fields and field values, e.g. name='foo'

        """
        activity = Activity()
        for key, value in self.items():
            if key != "id":
                activity[key] = value
        for k, v in kwargs.items():
            activity._data[k] = v
        activity._data["code"] = str(code or uuid.uuid4().hex)
        activity.save()

        for exc in self.exchanges():
            data = copy.deepcopy(exc._data)
            data["output"] = activity.key
            # Change `input` for production exchanges
            if exc["input"] == exc["output"]:
                data["input"] = activity.key
            ExchangeDataset.create(**dict_as_exchangedataset(data))
        return activity


class Exchange(ExchangeProxyBase):
    def __init__(self, document=None, **kwargs):
        """Create an `Exchange` proxy object.

        If this is a new exchange, can pass `kwargs`.

        If the exchange exists in the database, `document` should be an `ExchangeDataset`."""
        if document is None:
            self._document = ExchangeDataset()
            self._data = kwargs
        else:
            self._document = document
            self._data = self._document.data
            self._data["input"] = (
                self._document.input_database,
                self._document.input_code,
            )
            self._data["output"] = (
                self._document.output_database,
                self._document.output_code,
            )

    @writable_project
    def save(self):
        if not self.valid():
            raise ValidityError(
                "This exchange can't be saved for the "
                "following reasons\n\t* " + "\n\t* ".join(self.valid(why=True)[1])
            )

        databases.set_dirty(self["output"][0])

        for key, value in dict_as_exchangedataset(self._data).items():
            setattr(self._document, key, value)
        self._document.save()

    @writable_project
    def delete(self):
        from ..parameters import ParameterizedExchange

        ParameterizedExchange.delete().where(
            ParameterizedExchange.exchange == self._document.id
        ).execute()
        self._document.delete_instance()
        databases.set_dirty(self["output"][0])
        self = None



from bw2data.tests import bw2test
from bw2data import projects
from bw2data.backends import sqlite3_lci_db as db
from bw2data.database import DatabaseChooser
from copy import copy


@bw2test
def test_switch_project_correctly_switches_database_objects():
    database = DatabaseChooser("testy")
    data = {
        ("testy", "A"): {},
        ("testy", "C"): {"type": "biosphere"},
        ("testy", "B"): {
            "exchanges": [
                {"input": ("testy", "A"), "amount": 1, "type": "technosphere"},
                {"input": ("testy", "B"), "amount": 1, "type": "production"},
                {"input": ("testy", "C"), "amount": 1, "type": "biosphere"},
            ]
        },
    }
    database.write(data)

    table = db._tables[0]
    current_db_location = copy(db.db.database)
    assert table.select().count()

    projects.set_current("new one")
    assert not table.select().count()
    assert current_db_location != db.db.database



from .project import writable_project
from .serialization import SerializedDict, PickledDict, CompoundJSONDict
import datetime


class GeoMapping(PickledDict):
    """A dictionary that maps location codes to integers. Needed because parameter arrays have integer ``geo`` fields.

    File data is stored in ``geomapping.pickle``.

    This dictionary does not support setting items directly; instead, use the ``add`` method to add multiple keys."""

    filename = "geomapping.pickle"

    def __init__(self, *args, **kwargs):
        super(GeoMapping, self).__init__(*args, **kwargs)
        # At a minimum, "GLO" should always be present
        if "GLO" not in self:
            self.add(["GLO"])

    @writable_project
    def add(self, keys):
        """Add a set of keys. These keys can already be in the mapping; only new keys will be added.

        Args:
            * *keys* (list): The keys to add.

        """
        index = max(self.data.values()) if self.data else 0
        for i, key in enumerate(keys):
            if key not in self.data:
                self.data[key] = index + i + 1
        self.flush()

    def delete(self, keys):
        """Delete a set of keys.

        Args:
            *keys* (list): The keys to delete.

        """
        for key in keys:
            del self.data[key]
        self.flush()

    def __setitem__(self, key, value):
        raise NotImplementedError

    def __str__(self):
        return "Mapping from databases and methods to parameter indices."

    def __len__(self):
        return len(self.data)


class Databases(SerializedDict):
    """A dictionary for database metadata. This class includes methods to manage database versions. File data is saved in ``databases.json``."""

    filename = "databases.json"

    @writable_project
    def increment_version(self, database, number=None):
        """Increment the ``database`` version. Returns the new version."""
        self.data[database]["version"] += 1
        if number is not None:
            self.data[database]["number"] = number
        self.flush()
        return self.data[database]["version"]

    def version(self, database):
        """Return the ``database`` version"""
        return self.data[database].get("version")

    @writable_project
    def set_modified(self, database):
        self[database]["modified"] = datetime.datetime.now().isoformat()
        self.flush()

    @writable_project
    def set_dirty(self, database):
        self.set_modified(database)
        if self[database].get("dirty"):
            pass
        else:
            self[database]["dirty"] = True
            self.flush()

    def clean(self):
        from . import Database

        @writable_project
        def _clean():
            for x in self:
                if self[x].get("dirty"):
                    Database(x).process()
                    del self[x]["dirty"]
            self.flush()

        if not any(self[x].get("dirty") for x in self):
            return
        else:
            return _clean()

    @writable_project
    def __delitem__(self, name):
        from . import Database

        try:
            Database(name).delete(warn=False)
        except:
            pass

        super(Databases, self).__delitem__(name)


class CalculationSetups(PickledDict):
    """A dictionary for calculation setups.

    Keys:
    * `inv`: List of functional units, e.g. ``[{(key): amount}, {(key): amount}]``
    * `ia`: List of LCIA methods, e.g. ``[(method), (method)]``.

    """

    filename = "setups.pickle"


class DynamicCalculationSetups(PickledDict):
    """A dictionary for Dynamic calculation setups.

    Keys:
    * `inv`: List of functional units, e.g. ``[{(key): amount}, {(key): amount}]``
    * `ia`: Dictionary of orst case LCIA method and the relative dynamic LCIA method, e.g. `` [{dLCIA_method_1_worstcase:dLCIA_method_1 , dLCIA_method_2_worstcase:dLCIA_method_2}]``.

    """

    filename = "dynamicsetups.pickle"


class Methods(CompoundJSONDict):
    """A dictionary for method metadata. File data is saved in ``methods.json``."""

    filename = "methods.json"


class WeightingMeta(Methods):
    """A dictionary for weighting metadata. File data is saved in ``methods.json``."""

    filename = "weightings.json"


class NormalizationMeta(Methods):
    """A dictionary for normalization metadata. File data is saved in ``methods.json``."""

    filename = "normalizations.json"


class Preferences(PickledDict):
    """A dictionary of project-specific preferences."""

    filename = "preferences.pickle"

    def __init__(self, *args, **kwargs):
        super(Preferences, self).__init__(*args, **kwargs)

        # Default preferences
        if "use_cache" not in self:
            self["use_cache"] = True


databases = Databases()
geomapping = GeoMapping()
methods = Methods()
normalizations = NormalizationMeta()
preferences = Preferences()
weightings = WeightingMeta()
calculation_setups = CalculationSetups()
dynamic_calculation_setups = DynamicCalculationSetups()



from peewee import SqliteDatabase, BlobField
import pickle


class PickleField(BlobField):
    def db_value(self, value):
        return super(PickleField, self).db_value(pickle.dumps(value, protocol=4))

    def python_value(self, value):
        return pickle.loads(bytes(value))


class SubstitutableDatabase(object):
    def __init__(self, filepath, tables):
        self._filepath = filepath
        self._tables = tables
        self._database = self._create_database()

    def _create_database(self):
        db = SqliteDatabase(self._filepath)
        for model in self._tables:
            model.bind(db, bind_refs=False, bind_backrefs=False)
        db.connect()
        db.create_tables(self._tables)
        return db

    @property
    def db(self):
        return self._database

    def change_path(self, filepath):
        self.db.close()
        self._filepath = filepath
        self._database = self._create_database()

    def atomic(self):
        return self.db.atomic()

    def execute_sql(self, *args, **kwargs):
        return self.db.execute_sql(*args, **kwargs)

    def transaction(self):
        return self.db.transaction()

    def vacuum(self):
        print("Vacuuming database ")
        self.execute_sql("VACUUM;")



from bw2data.tests import BW2DataTest
from bw2data import projects
from bw2data.database import DatabaseChooser
from bw2data.backends import (
    Activity as PWActivity,
    ActivityDataset,
    Exchange as PWExchange,
    ExchangeDataset,
)
from bw2data.backends.utils import convert_backend
from bw2data.errors import (
    InvalidExchange,
    MissingIntermediateData,
    UnknownObject,
    UntypedExchange,
    ValidityError,
)
from bw2data.meta import geomapping, databases, methods


class DatabaseQuerysetTest(BW2DataTest):
    def extra_setup(self):
        self.db = DatabaseChooser("Order!")
        self.db.write(
            {
                ("Order!", "first"): {
                    "name": "a",
                    "location": "delaware",
                    "reference product": "widget",
                },
                ("Order!", "second"): {
                    "name": "b",
                    "location": "carolina",
                    "reference product": "wiggle",
                },
                ("Order!", "third"): {
                    "name": "c",
                    "location": "baseball",
                    "reference product": "lollipop",
                },
                ("Order!", "fourth"): {
                    "name": "d",
                    "location": "alabama",
                    "reference product": "widget",
                },
            }
        )

    def test_setup_clean(self):
        self.assertEqual(len(databases), 1)
        self.assertTrue("Order!" in databases)
        self.assertEqual(list(methods), [])
        self.assertEqual(len(geomapping), 5)  # GLO
        self.assertTrue("GLO" in geomapping)
        self.assertTrue("carolina" in geomapping)
        self.assertEqual(len(projects), 1)  # Default project
        self.assertTrue("default" in projects)

    def test_random_with_global_filters(self):
        self.db.filters = {"product": "lollipop"}
        for _ in range(10):
            self.assertEqual(self.db.random()["name"], "c")

    def test_random_with_local_filters(self):
        for _ in range(10):
            self.assertEqual(
                self.db.random(filters={"product": "lollipop"})["name"], "c"
            )

    def test_random_with_local_and_global_filters(self):
        db = DatabaseChooser("Newt")
        db.write(
            {
                ("Newt", "first"): {
                    "name": "a",
                    "location": "delaware",
                    "reference product": "widget",
                },
                ("Newt", "second"): {
                    "name": "b",
                    "location": "delaware",
                    "reference product": "wiggle",
                },
                ("Newt", "third"): {
                    "name": "c",
                    "location": "alabama",
                    "reference product": "widget",
                },
                ("Newt", "fourth"): {
                    "name": "d",
                    "location": "alabama",
                    "reference product": "wiggle",
                },
            }
        )
        self.assertTrue(len({db.random()["name"] for _ in range(10)}) > 1)
        db.filters = {"product": "widget"}
        for _ in range(10):
            self.assertEqual(
                self.db.random(filters={"location": "delaware"})["name"], "a"
            )

    def test_contains_respects_filters(self):
        self.db.filters = {"product": "lollipop"}
        self.assertFalse(("Order!", "fourth") in self.db)

    def test_get_ignores_filters(self):
        self.db.filters = {"product": "giggles"}
        self.assertEqual(self.db.get("fourth")["name"], "d")

    def test_filter(self):
        self.db.filters = {"product": "widget"}
        self.assertEqual(len([x for x in self.db]), 2)

    def test_order_by(self):
        self.db.order_by = "name"
        self.assertEqual([x["name"] for x in self.db], ["a", "b", "c", "d"])

    def test_order_by_bad_field(self):
        with self.assertRaises(AssertionError):
            self.db.order_by = "poopy"

    def test_filter_bad_field(self):
        with self.assertRaises(AssertionError):
            self.db.filters = {"poopy": "yuck"}

    def test_filter_not_dict(self):
        with self.assertRaises(AssertionError):
            self.db.filters = "poopy"

    def test_reset_order_by(self):
        self.db.order_by = "name"
        self.db.order_by = None
        as_lists = [[x["name"] for x in self.db] for _ in range(10)]
        first_elements = {x[0] for x in as_lists}
        self.assertTrue(len(first_elements) > 1)

    def test_reset_filters(self):
        self.db.filters = {"product": "widget"}
        self.assertEqual(len([x for x in self.db]), 2)
        self.db.filters = None
        self.assertEqual(len([x for x in self.db]), 4)

    def test_len_respects_filters(self):
        self.db.filters = {"product": "widget"}
        self.assertEqual(len(self.db), 2)

    def test_make_searchable_unknown_object(self):
        db = DatabaseChooser("mysterious")
        with self.assertRaises(UnknownObject):
            db.make_searchable()

    def test_convert_same_backend(self):
        database = DatabaseChooser("a database")
        database.write(
            {
                ("a database", "foo"): {
                    "exchanges": [
                        {
                            "input": ("a database", "foo"),
                            "amount": 1,
                            "type": "production",
                        }
                    ],
                    "location": "bar",
                    "name": "baz",
                },
            }
        )
        self.assertFalse(convert_backend("a database", "sqlite"))

    def test_convert_backend(self):
        self.maxDiff = None
        database = DatabaseChooser("a database")
        database.write(
            {
                ("a database", "foo"): {
                    "exchanges": [
                        {
                            "input": ("a database", "foo"),
                            "amount": 1,
                            "type": "production",
                        }
                    ],
                    "location": "bar",
                    "name": "baz",
                },
            }
        )



from .. import config
from ..meta import methods, databases
from ..method import Method
from ..errors import UntypedExchange, InvalidExchange
import copy
import numpy as np
import warnings


def get_csv_data_dict(ds):
    fields = {"name", "reference product", "unit", "location"}
    dd = {field: ds.get(field) for field in fields}
    dd["id"] = get_id(ds)
    return dd


def convert_backend(database_name, backend):
    """Convert a Database to another backend.

    bw2data currently supports the `default` and `json` backends.

    Args:
        * `database_name` (unicode): Name of database.
        * `backend` (unicode): Type of database. `backend` should be recoginized by `DatabaseChooser`.

    Returns `False` if the old and new backend are the same. Otherwise returns an instance of the new Database object."""
    if database_name not in databases:
        print("Can't find database {}".format(database_name))

    from ..database import Database

    db = Database(database_name)
    if db.backend == backend:
        return False
    # Needed to convert from async json dict
    data = db.load(as_dict=True)
    if database_name in config.cache:
        del config.cache[database_name]
    metadata = copy.deepcopy(db.metadata)
    metadata["backend"] = str(backend)
    del databases[database_name]
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        new_db = Database(database_name, backend)
        new_db.register(**metadata)
    new_db.write(data)
    return new_db


def check_exchange(exc):
    """Check exchange data validity when processing"""
    if "type" not in exc:
        raise UntypedExchange
    if "amount" not in exc or "input" not in exc:
        raise InvalidExchange
    if np.isnan(exc["amount"]) or np.isinf(exc["amount"]):
        raise ValueError("Invalid amount in exchange {}".format(exc))


def dict_as_activitydataset(ds):
    return {
        "data": ds,
        "database": ds["database"],
        "code": ds["code"],
        "location": ds.get("location"),
        "name": ds.get("name"),
        "product": ds.get("reference product"),
        "type": ds.get("type", "process"),
    }


def dict_as_exchangedataset(ds):
    return {
        "data": ds,
        "input_database": ds["input"][0],
        "input_code": ds["input"][1],
        "output_database": ds["output"][0],
        "output_code": ds["output"][1],
        "type": ds["type"],
    }


def replace_cfs(old_key, new_key):
    """Replace ``old_key`` with ``new_key`` in characterization factors.

    Returns list of modified methods."""
    altered_methods = []
    for name in methods:
        changed = False
        data = Method(name).load()
        for line in data:
            if line[0] == old_key:
                line[0], changed = new_key, True
        if changed:
            Method(name).write(data)
            altered_methods.append(name)
    return altered_methods


def retupleize_geo_strings(value):
    """Transform data from SQLite representation to Python objects.

    We are using a SQLite3 cursor, which means that the Peewee data conversion code is not called. So ``('foo', 'bar')`` is stored as a string, not a tuple. This code tries to do this conversion correctly.

    TODO: Adapt what Peewee does in this case?"""
    if not value:
        return value
    elif "(" not in value:
        return value
    try:
        # Is this a dirty, dirty hack, or inspiration?
        # Location is retrieved as a string from the database
        # The alternative is to retrieve and process the
        # entire activity dataset...
        return eval(value)
    except NameError:
        # Not everything with a parentheses is a tuple.
        return value



from . import projects
from .errors import UnknownObject, MissingIntermediateData
from .fatomic import open as atomic_open
from .project import writable_project
from bw_processing import create_datapackage, clean_datapackage_name, safe_filename
import pickle
from fs.zipfs import ZipFS


class DataStore(object):
    """
Base class for all Brightway2 data stores. Subclasses should define:

    * **metadata**: A :ref:`serialized-dict` instance, e.g. ``databases`` or ``methods``. The custom is that each type of data store has a new metadata store, so the data store ``Foo`` would have a metadata store ``foos``.
    * **validator**: A data validator. Optional. See bw2data.validate.

    """

    validator = None
    _metadata = None
    _intermediate_dir = "intermediate"

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return "Brightway2 %s: %s" % (self.__class__.__name__, self.name)

    __repr__ = lambda self: str(self)

    def _get_metadata(self):
        if self.name not in self._metadata:
            raise UnknownObject(
                "This object is not yet registered; can't get or set metadata"
            )
        return self._metadata[self.name]

    @writable_project
    def _set_metadata(self, value):
        self._get_metadata()
        self._metadata[self.name] = value
        self._metadata.flush()

    metadata = property(_get_metadata, _set_metadata)

    @property
    def filename(self):
        """Remove filesystem-unsafe characters and perform unicode normalization on ``self.name`` using :func:`.filesystem.safe_filename`."""
        return safe_filename(self.name)

    @property
    def registered(self):
        return self.name in self._metadata

    def register(self, **kwargs):
        """Register an object with the metadata store. Takes any number of keyword arguments."""

        @writable_project
        def _register(kwargs):
            self._metadata[self.name] = kwargs

        if self.name not in self._metadata:
            _register(kwargs)

    @writable_project
    def deregister(self):
        """Remove an object from the metadata store. Does not delete any files."""
        del self._metadata[self.name]

    def load(self):
        """Load the intermediate data for this object.

        Returns:
            The intermediate data.

        """
        if not self.registered:
            raise UnknownObject("This object is not registered and has no data")
        try:
            return pickle.load(
                open(projects.dir / "intermediate" / (self.filename + ".pickle"), "rb",)
            )
        except OSError:
            raise MissingIntermediateData("Can't load intermediate data")

    @writable_project
    def copy(self, name):
        """Make a copy of this object with a new ``name``.

        This method only changes the name, but not any of the data or metadata.

        Args:
            * *name* (object): Name of the new object.

        Returns:
            The new object.

        """
        assert name not in self._metadata, "%s already exists" % name
        new_obj = self.__class__(name)
        new_obj.register(**self.metadata)
        new_obj.write(self.load())
        return new_obj

    def backup(self):
        """Save a backup to ``backups`` folder.

        Returns:
            File path of backup.

        """
        try:
            from bw2io import BW2Package

            return BW2Package.export_obj(self)
        except ImportError:
            print("bw2io not installed")

    @writable_project
    def write(self, data):
        """Serialize intermediate data to disk.

        Args:
            * *data* (object): The data

        """
        self.register()
        filepath = projects.dir / self._intermediate_dir / (self.filename + ".pickle")
        with atomic_open(filepath, "wb") as f:
            pickle.dump(data, f, protocol=4)

    def validate(self, data):
        """Validate data. Must be called manually."""
        self.validator(data)
        return True


class ProcessedDataStore(DataStore):
    """
Brightway2 data stores that can be processed to NumPy arrays.

In addition to ``metadata`` and (optionally) ``validator``, subclasses should override ``add_geomappings``. This method takes the entire dataset, and loads objects to :ref:`geomapping` as needed.

    """

    matrix = "unknown"

    def dirpath_processed(self):
        return projects.dir / "processed"

    def filename_processed(self):
        return clean_datapackage_name(self.filename + ".zip")

    def filepath_processed(self):
        return self.dirpath_processed() / self.filename_processed()

    @writable_project
    def write(self, data, process=True):
        """Serialize intermediate data to disk.

        Args:
            * *data* (object): The data

        """
        self.register()
        self.add_geomappings(data)
        filepath = projects.dir / self._intermediate_dir / (self.filename + ".pickle")
        with atomic_open(filepath, "wb") as f:
            pickle.dump(data, f, protocol=4)
        if process:
            self.process()

    def process_row(self, row):
        """Translate data into a dictionary suitable for array inputs.

        See `bw_processing documentation <https://github.com/brightway-lca/bw_processing>`__."""
        raise NotImplementedError

    def process(self, **extra_metadata):
        """
Process intermediate data from a Python dictionary to a `stats_arrays <https://pypi.python.org/pypi/stats_arrays/>`_ array, which is a `NumPy <http://numpy.scipy.org/>`_ `Structured <http://docs.scipy.org/doc/numpy/reference/generated/numpy.recarray.html#numpy.recarray>`_ `Array <http://docs.scipy.org/doc/numpy/user/basics.rec.html>`_. A structured array (also called record array) is a heterogeneous array, where each column has a different label and data type.

Processed arrays are saved in the ``processed`` directory.

If the uncertainty type is no uncertainty, undefined, or not specified, then the 'amount' value is used for 'loc' as well. This is needed for the random number generator.

Doesn't return anything, but writes a file to disk.

        """
        data = self.load()
        dp = create_datapackage(
            fs=ZipFS(str(self.filepath_processed()), write=True),
            name=self.filename_processed(),
            sum_intra_duplicates=True,
            sum_inter_duplicates=False,
        )
        dp.add_persistent_vector_from_iterator(
            matrix=self.matrix,
            name=clean_datapackage_name(str(self.name) + " matrix data"),
            dict_iterator=(self.process_row(row) for row in data),
            nrows=len(data),
            **extra_metadata
        )
        dp.finalize_serialization()

    def add_geomappings(self, data):
        """Add objects to ``geomapping``, if necessary.

        Args:
            * *data* (object): The data

        """
        return

    def validate(self, data):
        """Validate data. Must be called manually."""
        self.validator(data)
        return True



from .. import SQLiteBackend
from ... import geomapping, config, databases
from bw_processing import create_datapackage, clean_datapackage_name
from fs.zipfs import ZipFS
import datetime


class IOTableBackend(SQLiteBackend):
    """IO tables have too much data to store each value in a database; instead, we only store the processed data in NumPy arrays.

    Activities will not seem to have any activities."""

    backend = "iotable"

    def write(self, data):
        super().write(data, process=False)

    def write_exchanges(self, technosphere, biosphere, dependents):
        """

        Write IO data directly to processed arrays.

        Product data is stored in SQLite as normal activities.
        Exchange data is written directly to NumPy structured arrays.

        Technosphere and biosphere data has format ``(row id, col id, value, flip)``.

        """
        print("Starting IO table write")

        dp = create_datapackage(
            fs=ZipFS(str(self.filepath_processed()), write=True),
            name=clean_datapackage_name(self.name),
            sum_intra_duplicates=True,
            sum_inter_duplicates=False,
        )

        dp.add_persistent_vector_from_iterator(
            dict_iterator=(
                {
                    "row": obj.id,
                    "col": geomapping[obj["location"] or config.global_location],
                    "amount": 1,
                }
                for obj in self
            ),
            matrix="inv_geomapping_matrix",
            name=clean_datapackage_name(self.name + " inventory geomapping matrix"),
            nrows=len(self),
        )
        print("Adding technosphere matrix")
        dp.add_persistent_vector_from_iterator(
            matrix="technosphere_matrix",
            name=clean_datapackage_name(self.name + " technosphere matrix"),
            dict_iterator=technosphere,
        )

        print("Adding biosphere matrix")
        dp.add_persistent_vector_from_iterator(
            matrix="biosphere_matrix",
            name=clean_datapackage_name(self.name + " biosphere matrix"),
            dict_iterator=biosphere,
        )
        dp.finalize_serialization()

        databases[self.name]["depends"] = sorted(set(dependents).difference({self.name}))
        databases[self.name]["processed"] = datetime.datetime.now().isoformat()
        databases.flush()

    def process(self):
        """No-op; no intermediate data to process"""
        return



from voluptuous import Schema, Required, Invalid, Any, All, Length, Optional
from numbers import Number


def valid_tuple(obj):
    try:
        assert isinstance(obj, tuple)
        assert isinstance(obj[0], str)
        assert isinstance(obj[1], str)
    except:
        raise Invalid("{} is not a valid key tuple".format(obj))
    return obj


uncertainty_dict = {
    Required("amount"): Number,
    Optional("uncertainty type"): int,
    Optional("loc"): Number,
    Optional("scale"): Number,
    Optional("shape"): Number,
    Optional("minimum"): Number,
    Optional("maximum"): Number,
}

exchange = {
    Required("input"): valid_tuple,
    Required("type"): str,
}
exchange.update(uncertainty_dict)

lci_dataset = {
    Optional("categories"): Any(list, tuple),
    Optional("location"): object,
    Optional("unit"): str,
    Optional("name"): str,
    Optional("type"): str,
    Optional("exchanges"): [exchange],
}

db_validator = Schema({valid_tuple: lci_dataset}, extra=True)

# TODO: elements in a list don't maintain order
# See https://github.com/alecthomas/voluptuous/issues/59
# Each list needs to be a separate function...

maybe_uncertainty = Any(Number, uncertainty_dict)

ia_validator = Schema(
    [
        Any(
            [valid_tuple, maybe_uncertainty],  # site-generic
            [valid_tuple, maybe_uncertainty, object],  # regionalized
        )
    ]
)

weighting_validator = Schema(All([uncertainty_dict], Length(min=1, max=1)))

normalization_validator = Schema([[valid_tuple, maybe_uncertainty]])



import platform


class Config(object):
    """A singleton that stores configuration settings"""

    version = 3
    backends = {}
    cache = {}
    metadata = []
    sqlite3_databases = []
    _windows = platform.system() == "Windows"

    @property
    def biosphere(self):
        """Get name for ``biosphere`` database from user preferences.

        Default name is ``biosphere3``; change this by changing ``config.p["biosphere_database"]``."""
        return self.p.get("biosphere_database", "biosphere3")

    @property
    def global_location(self):
        """Get name for global location from user preferences.

        Default name is ``GLO``; change this by changing ``config.p["global_location"]``."""
        return self.p.get("global_location", "GLO")


config = Config()



from bw2data.tests import bw2test
from bw2data import config


@bw2test
def test_default_biosphere():
    assert config.biosphere == "biosphere3"


@bw2test
def test_default_geo():
    assert config.global_location == "GLO"



from ..errors import UnknownObject
from ..sqlite import PickleField
from functools import lru_cache
from peewee import Model, TextField, DoesNotExist


class ActivityDataset(Model):
    data = PickleField()  # Canonical, except for other C fields
    code = TextField()  # Canonical
    database = TextField()  # Canonical
    location = TextField(null=True)  # Reset from `data`
    name = TextField(null=True)  # Reset from `data`
    product = TextField(null=True)  # Reset from `data`
    type = TextField(null=True)  # Reset from `data`

    @property
    def key(self):
        return (self.database, self.code)


class ExchangeDataset(Model):
    data = PickleField()  # Canonical, except for other C fields
    input_code = TextField()  # Canonical
    input_database = TextField()  # Canonical
    output_code = TextField()  # Canonical
    output_database = TextField()  # Canonical
    type = TextField()  # Reset from `data`


@lru_cache(maxsize=4096)
def get_id(key):
    if isinstance(key, int):
        return key
    else:
        try:
            return ActivityDataset.get(
                ActivityDataset.database == key[0], ActivityDataset.code == key[1]
            ).id
        except DoesNotExist:
            raise UnknownObject



from bw2data.tests import BW2DataTest, bw2test
from bw2data import *
from .fixtures import food, biosphere


@bw2test
def test_repr_str_unicode():
    objects = (
        geomapping,
        databases,
        methods,
        normalizations,
        weightings,
        Database("foo"),
        DataStore("foo"),
        projects,
    )
    for obj in objects:
        assert repr(obj)
        assert str(obj)
        print(obj)


@bw2test
def test_registered_database_repr():
    d = Database("biosphere")
    d.write(biosphere)
    assert repr(d)
    assert str(d)
    # Make sure can be printed - not for debugging
    print(d)



from . import databases, projects, config, get_activity
from .backends.schema import ExchangeDataset
from .sqlite import PickleField, SubstitutableDatabase
from asteval import Interpreter
from bw2parameters import ParameterSet
from bw2parameters.errors import MissingName
from peewee import (
    BooleanField,
    Check,
    DateTimeField,
    FloatField,
    IntegerField,
    Model,
    TextField,
)
import asteval
import datetime
import itertools
import re
import uuid


# https://stackoverflow.com/questions/34544784/arbitrary-string-to-valid-python-name
clean = lambda x: re.sub(r"\W|^(?=\d)", "_", x)
nonempty = lambda dct: {k: v for k, v in dct.items() if v is not None}

"""Autoupdate `updated` field in Group when parameters change"""
AUTOUPDATE_TRIGGER = """CREATE TRIGGER IF NOT EXISTS {table}_{action}_trigger AFTER {action} ON {table} BEGIN
    UPDATE group_table SET updated = datetime('now') WHERE name = {name};
END;"""

"""Activity parameter groups can't cross databases"""
_CROSSDATABASE_TEMPLATE = """CREATE TRIGGER IF NOT EXISTS ap_crossdatabase_{action} BEFORE {action} ON activityparameter BEGIN
    SELECT CASE WHEN
        ((SELECT COUNT(*) FROM activityparameter WHERE "group" = NEW."group") > 0)
    AND (NEW.database NOT IN (SELECT DISTINCT "database" FROM activityparameter where "group" = NEW."group"))
    THEN RAISE(ABORT,'Cross database group')
    END;
END;"""
CROSSDATASE_INSERT_TRIGGER = _CROSSDATABASE_TEMPLATE.format(action="INSERT")
CROSSDATASE_UPDATE_TRIGGER = _CROSSDATABASE_TEMPLATE.format(action="UPDATE")

"""Activities can't be in multiple activity parameter groups"""
_CROSSGROUP_TEMPLATE = """CREATE TRIGGER IF NOT EXISTS ap_crossgroup_{action} BEFORE {action} ON activityparameter BEGIN
    SELECT CASE WHEN EXISTS (SELECT * FROM activityparameter AS a WHERE
            a.database = NEW.database AND
            a.code = NEW.code AND
            a."group" != NEW."group")
    THEN RAISE(ABORT,'Cross group activity')
    END;
END;"""
CROSSGROUP_INSERT_TRIGGER = _CROSSGROUP_TEMPLATE.format(action="INSERT")
CROSSGROUP_UPDATE_TRIGGER = _CROSSGROUP_TEMPLATE.format(action="UPDATE")

"""No circular dependences in activity parameter group dependencies"""
_CLOSURE_TEMPLATE = """CREATE TRIGGER IF NOT EXISTS gd_circular_{action} BEFORE {action} ON groupdependency BEGIN
    SELECT CASE WHEN EXISTS (SELECT * FROM groupdependency AS g WHERE g."group" = NEW.depends AND g.depends = NEW."group")
    THEN RAISE(ABORT,'Circular dependency')
    END;
END;
"""
GD_INSERT_TRIGGER = _CLOSURE_TEMPLATE.format(action="INSERT")
GD_UPDATE_TRIGGER = _CLOSURE_TEMPLATE.format(action="UPDATE")

"""Parameterized exchange groups must be in activityparameters table"""
_PE_GROUP_TEMPLATE = """CREATE TRIGGER IF NOT EXISTS pe_group_{action} BEFORE {action} ON parameterizedexchange BEGIN
    SELECT CASE WHEN
        ((SELECT COUNT(*) FROM activityparameter WHERE "group" = NEW."group") < 1)
    THEN RAISE(ABORT,'Missing activity parameter group')
    END;
END;
"""
PE_INSERT_TRIGGER = _PE_GROUP_TEMPLATE.format(action="INSERT")
PE_UPDATE_TRIGGER = _PE_GROUP_TEMPLATE.format(action="UPDATE")


class ParameterBase(Model):
    __repr__ = lambda x: str(x)

    def __lt__(self, other):
        if type(self) != type(other):
            raise TypeError
        else:
            return self.name.lower() < other.name.lower()

    @classmethod
    def create_table(cls):
        super(ParameterBase, cls).create_table()
        cls._meta.database.execute_sql(
            AUTOUPDATE_TRIGGER.format(
                action="INSERT", name=cls._new_name, table=cls._db_table
            )
        )
        for action in ("UPDATE", "DELETE"):
            cls._meta.database.execute_sql(
                AUTOUPDATE_TRIGGER.format(
                    action=action, name=cls._old_name, table=cls._db_table
                )
            )

    @staticmethod
    def expire_downstream(group):
        """Expire any activity parameters that depend on this group"""
        Group.update(fresh=False).where(
            Group.name
            << GroupDependency.select(GroupDependency.group).where(
                GroupDependency.depends == group
            )
        ).execute()


class ProjectParameter(ParameterBase):
    """Parameter set for a project. Group name is 'project'.

    Columns:

        * name: str, unique
        * formula: str, optional
        * amount: float, optional
        * data: object, optional. Used for any other metadata.

    Note that there is no magic for reading and writing to ``data`` (unlike ``Activity`` objects) - it must be used directly.

    """

    name = TextField(index=True, unique=True)
    formula = TextField(null=True)
    amount = FloatField(null=True)
    data = PickleField(default={})

    _old_name = "'project'"
    _new_name = "'project'"
    _db_table = "projectparameter"

    def __str__(self):
        return "Project parameter: {}".format(self.name)

    def save(self, *args, **kwargs):
        Group.get_or_create(name="project")[0].expire()
        super(ProjectParameter, self).save(*args, **kwargs)

    @staticmethod
    def load(group=None):
        """Return dictionary of parameter data with names as keys and ``.dict()`` as values."""

        def reformat(o):
            o = o.dict
            return (o.pop("name"), o)

        return dict([reformat(o) for o in ProjectParameter.select()])

    @staticmethod
    def static(ignored="project", only=None):
        """Get dictionary of ``{name: amount}`` for all project parameters.

        ``only`` restricts returned names to ones found in ``only``. ``ignored`` included for API compatibility with other ``recalculate`` methods."""
        result = dict(
            ProjectParameter.select(
                ProjectParameter.name, ProjectParameter.amount
            ).tuples()
        )
        if only is not None:
            result = {k: v for k, v in result.items() if k in only}
        return result

    @staticmethod
    def expired():
        """Return boolean - is this group expired?"""
        try:
            return not Group.get(name="project").fresh
        except Group.DoesNotExist:
            return False

    @staticmethod
    def recalculate(ignored=None):
        """Recalculate all parameters.

        ``ignored`` included for API compatibility with other ``recalculate`` methods - it will really be ignored."""
        if not ProjectParameter.expired():
            return
        data = ProjectParameter.load()
        if not data:
            return
        ParameterSet(data).evaluate_and_set_amount_field()
        with parameters.db.atomic() as _:
            for key, value in data.items():
                ProjectParameter.update(amount=value["amount"],).where(
                    ProjectParameter.name == key
                ).execute()
            Group.get_or_create(name="project")[0].freshen()
            ProjectParameter.expire_downstream("project")

    @staticmethod
    def dependency_chain():
        """ Determine if ```ProjectParameter`` parameters have dependencies
        within the group.

        Returns:

        .. code-block:: python

            [
                {
                    'kind': 'project',
                    'group': 'project',
                    'names': set of variables names
                }
            ]

        """
        data = ProjectParameter.load()
        if not data:
            return []

        # Parse all formulas, find missing variables
        needed = get_new_symbols(data.values())
        if not needed:
            return []

        missing = needed.difference(data)
        if missing:
            raise MissingName(
                "The following variables aren't defined:\n{}".format("|".join(missing))
            )

        return [{"kind": "project", "group": "project", "names": needed}]

    @staticmethod
    def is_dependency_within_group(name):
        own_group = next(iter(ProjectParameter.dependency_chain()), {})
        return True if name in own_group.get("names", set()) else False

    def is_deletable(self):
        """Perform a test to see if the current parameter can be deleted."""
        if ProjectParameter.is_dependency_within_group(self.name):
            return False
        # Test the database parameters
        if DatabaseParameter.is_dependent_on(self.name):
            return False
        # Test activity parameters
        if ActivityParameter.is_dependent_on(self.name, "project"):
            return False
        return True

    @classmethod
    def update_formula_parameter_name(cls, old, new):
        """ Performs an update of the formula of relevant parameters.

        NOTE: Make sure to wrap this in an .atomic() statement!
        """
        data = (
            alter_parameter_formula(p, old, new)
            for p in cls.select().where(cls.formula.contains(old))
        )
        cls.bulk_update(data, fields=[cls.formula], batch_size=50)
        Group.get_or_create(name="project")[0].expire()

    @property
    def dict(self):
        """Parameter data as a standardized dictionary"""
        obj = nonempty(
            {"name": self.name, "formula": self.formula, "amount": self.amount,}
        )
        obj.update(self.data)
        return obj


class DatabaseParameter(ParameterBase):
    """Parameter set for a database. Group name is the name of the database.

    Columns:

        * database: str
        * name: str, unique within a database
        * formula: str, optional
        * amount: float, optional
        * data: object, optional. Used for any other metadata.

    Note that there is no magic for reading and writing to ``data`` (unlike ``Activity`` objects) - it must be used directly.

    """

    database = TextField(index=True)
    name = TextField(index=True)
    formula = TextField(null=True)
    amount = FloatField(null=True)
    data = PickleField(default={})

    _old_name = "OLD.database"
    _new_name = "NEW.database"
    _db_table = "databaseparameter"

    class Meta:
        indexes = ((("database", "name"), True),)
        constraints = [Check("database != 'project'")]

    def __str__(self):
        return "Database parameter: {}:{}".format(self.database, self.name)

    @staticmethod
    def load(database):
        """Return dictionary of parameter data with names as keys and ``.dict()`` as values."""

        def reformat(o):
            o = o.dict
            return (o.pop("name"), o)

        return dict(
            [
                reformat(o)
                for o in DatabaseParameter.select().where(
                    DatabaseParameter.database == database
                )
            ]
        )

    @staticmethod
    def expired(database):
        """Return boolean - is this group expired?"""
        try:
            return not Group.get(name=database).fresh
        except Group.DoesNotExist:
            return False

    @staticmethod
    def static(database, only=None):
        """Return dictionary of {name: amount} for database group."""
        result = dict(
            DatabaseParameter.select(DatabaseParameter.name, DatabaseParameter.amount)
            .where(DatabaseParameter.database == database)
            .tuples()
        )
        if only is not None:
            result = {k: v for k, v in result.items() if k in only}
        return result

    @staticmethod
    def recalculate(database):
        """Recalculate all database parameters for ``database``, if expired."""
        if ProjectParameter.expired():
            ProjectParameter.recalculate()

        # Can we avoid doing anything?
        if not DatabaseParameter.expired(database):
            return
        data = DatabaseParameter.load(database)
        if not data:
            return

        # Parse all formulas, find missing variables
        new_symbols = get_new_symbols(data.values(), set(data))
        found_symbols = {
            x[0] for x in ProjectParameter.select(ProjectParameter.name).tuples()
        }
        missing = new_symbols.difference(found_symbols)
        if missing:
            raise MissingName(
                "The following variables aren't defined:\n{}".format("|".join(missing))
            )

        # Add or delete `project` dependency if needed
        if new_symbols:
            GroupDependency.get_or_create(group=database, depends="project")
            # Load needed project variables as {'foo': 42} dict
            glo = ProjectParameter.static(only=new_symbols)
        else:
            GroupDependency.delete().where(
                GroupDependency.group == database, GroupDependency.depends == "project"
            ).execute()
            glo = None

        # Update database parameter values
        ParameterSet(data, glo).evaluate_and_set_amount_field()
        with parameters.db.atomic():
            for key, value in data.items():
                DatabaseParameter.update(amount=value["amount"],).where(
                    DatabaseParameter.name == key,
                    DatabaseParameter.database == database,
                ).execute()
            Group.get(name=database).freshen()
            DatabaseParameter.expire_downstream(database)

    @staticmethod
    def dependency_chain(group, include_self=False):
        """Find where each missing variable is defined in dependency chain.

        If ``include_self`` is True will include parameters within the group as possible dependencies

        Returns:

        .. code-block:: python

            [
                {
                    'kind': one of 'project', 'database', 'activity',
                    'group': group name,
                    'names': set of variables names
                }
            ]

        """
        data = DatabaseParameter.load(group)
        if not data:
            return []

        # Parse all formulas, find missing variables
        context = set(data) if not include_self else set()
        needed = get_new_symbols(data.values(), context=context)
        if not needed:
            return []

        names, chain = set(), []
        if include_self:
            included = needed.intersection(data)
            for name in included:
                names.add(name)
                needed.remove(name)
            if names:
                chain.append({"kind": "database", "group": group, "names": names})

        if needed:
            names = set()
            for name in ProjectParameter.static(only=needed):
                names.add(name)
                needed.remove(name)
            if names:
                chain.insert(0, {"kind": "project", "group": "project", "names": names})

        if needed:
            raise MissingName(
                "The following variables aren't defined:\n{}".format("|".join(needed))
            )

        return chain

    @staticmethod
    def is_dependency_within_group(name, database):
        own_group = next(
            (
                x
                for x in DatabaseParameter.dependency_chain(database, include_self=True)
                if x.get("group") == database
            ),
            {},
        )
        return True if name in own_group.get("names", set()) else False

    def save(self, *args, **kwargs):
        """Save this model instance"""
        Group.get_or_create(name=self.database)[0].expire()
        super(DatabaseParameter, self).save(*args, **kwargs)

    def is_deletable(self):
        """Perform a test to see if the current parameter can be deleted."""
        # Test if the current parameter is used by other database parameters
        if DatabaseParameter.is_dependency_within_group(self.name, self.database):
            return False
        # Then test all relevant activity parameters
        if ActivityParameter.is_dependent_on(self.name, self.database):
            return False
        return True

    @staticmethod
    def is_dependent_on(name):
        """ Test if any database parameters are dependent on the given
        project parameter name.
        """
        query = (
            GroupDependency.select(GroupDependency.group)
            .where(GroupDependency.depends == "project")
            .distinct()
        )

        for row in query.execute():
            chain = DatabaseParameter.dependency_chain(row.group)
            own_group = next((x for x in chain if x.get("group") == "project"), {})
            if name in own_group.get("names", set()):
                return True

        return False

    @classmethod
    def update_formula_project_parameter_name(cls, old, new):
        """ Performs an update of the formula of relevant parameters.

        This method specifically targets project parameters used in database
        formulas
        """
        data = (
            alter_parameter_formula(p, old, new)
            for p in (
                cls.select()
                .join(GroupDependency, on=(GroupDependency.group == cls.database))
                .where(cls.formula.contains(old))
            )
            if not DatabaseParameter.is_dependency_within_group(old, p.database)
        )
        dbs = set(
            p.database
            for p in (
                cls.select(cls.database)
                .join(GroupDependency, on=(GroupDependency.group == cls.database))
                .where(cls.formula.contains(old))
                .distinct()
            )
            if not DatabaseParameter.is_dependency_within_group(old, p.database)
        )
        cls.bulk_update(data, fields=[cls.formula], batch_size=50)
        for db in dbs:
            Group.get_or_create(name=db)[0].expire()

    @classmethod
    def update_formula_database_parameter_name(cls, old, new):
        """ Performs an update of the formula of relevant parameters.

        This method specifically targets database parameters used in database
        formulas
        """
        data = (
            alter_parameter_formula(p, old, new)
            for p in cls.select().where(cls.formula.contains(old))
            if DatabaseParameter.is_dependency_within_group(old, p.database)
        )
        dbs = set(
            p.database
            for p in (
                cls.select(cls.database).where(cls.formula.contains(old)).distinct()
            )
            if DatabaseParameter.is_dependency_within_group(old, p.database)
        )
        cls.bulk_update(data, fields=[cls.formula], batch_size=50)
        for db in dbs:
            Group.get_or_create(name=db)[0].expire()

    @property
    def dict(self):
        """Parameter data as a standardized dictionary"""
        obj = nonempty(
            {
                "database": self.database,
                "name": self.name,
                "formula": self.formula,
                "amount": self.amount,
            }
        )
        obj.update(self.data)
        return obj


class ActivityParameter(ParameterBase):
    """Parameter set for a group of activities.

    Columns:

        * group: str
        * database: str
        * code: str. Code and database define the linked activity for this parameter.
        * name: str, unique within a group
        * formula: str, optional
        * amount: float, optional
        * data: object, optional. Used for any other metadata.

    Activities can only have parameters in one group. Group names cannot be 'project' or the name of any existing database.

    Activity parameter groups can depend on other activity parameter groups, so that a formula in group "a" can depend on a variable in group "b". This dependency information is stored in ``Group.order`` - in our small example, we could define the following:

    .. code-block:: python

        a = Group.get(name="a")
        a.order = ["b", "c"]
        a.save()

    In this case, a variable not found in "a" would be searched for in "b" and then "c", in that order. Database and then project parameters are also implicitly included at the end of ``Group.order``.

    Note that there is no magic for reading and writing to ``data`` (unlike ``Activity`` objects) - it must be used directly.

    """

    group = TextField()
    database = TextField()
    code = TextField()
    name = TextField()
    formula = TextField(null=True)
    amount = FloatField(null=True)
    data = PickleField(default={})

    _old_name = 'OLD."group"'
    _new_name = 'NEW."group"'
    _db_table = "activityparameter"

    class Meta:
        indexes = [(("group", "name"), True)]
        constraints = [Check("""("group" != 'project') AND ("group" != database)""")]

    def __str__(self):
        return "Activity parameter: {}:{}".format(self.group, self.name)

    @staticmethod
    def load(group):
        """Return dictionary of parameter data with names as keys and ``.dict()`` as values."""

        def reformat(o):
            o = o.dict
            return (o.pop("name"), o)

        return dict(
            [
                reformat(o)
                for o in ActivityParameter.select().where(
                    ActivityParameter.group == group
                )
            ]
        )

    @staticmethod
    def static(group, only=None, full=False):
        """Get dictionary of ``{name: amount}`` for parameters defined in ``group``.

        ``only`` restricts returned names to ones found in ``only``. ``full`` returns all names, including those found in the dependency chain."""
        result = dict(
            ActivityParameter.select(ActivityParameter.name, ActivityParameter.amount)
            .where(ActivityParameter.group == group)
            .tuples()
        )
        if full:
            temp = ActivityParameter._static_dependencies(group)
            temp.update(result)
            result = temp
        if only is not None:
            result = {k: v for k, v in result.items() if k in only}
        return result

    @staticmethod
    def _static_dependencies(group):
        """Get dictionary of ``{name: amount}`` for all variables defined in dependency chain.

        Be careful! This could have variables which overlap with local variable names. Designed for internal use."""
        database = ActivityParameter.get(group=group).database

        chain = [ProjectParameter.static(), DatabaseParameter.static(database)] + [
            ActivityParameter.static(g) for g in Group.get(name=group).order[::-1]
        ]

        result = {}
        for dct in chain:
            result.update(dct)
        return result

    @staticmethod
    def insert_dummy(group, activity):
        code, database = activity[1], activity[0]
        if (
            not ActivityParameter.select()
            .where(
                ActivityParameter.group == group,
                ActivityParameter.code == code,
                ActivityParameter.database == database,
            )
            .count()
        ):
            ActivityParameter.create(
                group=group,
                name="__dummy_{}__".format(uuid.uuid4().hex),
                code=code,
                database=database,
                amount=0,
            )

    @staticmethod
    def expired(group):
        """Return boolean - is this group expired?"""
        try:
            return not Group.get(name=group).fresh
        except Group.DoesNotExist:
            return False

    @staticmethod
    def dependency_chain(group, include_self=False):
        """Find where each missing variable is defined in dependency chain.

        Will also load in all parameters needed to resolve the ``ParameterizedExchanges`` for this group.

        If ``include_self`` is True will include parameters within the group as possible dependencies

        Returns:

        .. code-block:: python

            [
                {
                    'kind': one of 'project', 'database', 'activity',
                    'group': group name,
                    'names': set of variables names
                }
            ]

        """
        data = ActivityParameter.load(group)
        if not data:
            return []

        # Parse all formulas, find missing variables
        context = set(data) if not include_self else None
        activity_needed = get_new_symbols(data.values(), context=context)
        exchanges_needed = get_new_symbols(
            ParameterizedExchange.load(group).values(), context=context
        )
        needed = activity_needed.union(exchanges_needed)

        if not needed:
            return []

        chain = []

        # Iteratively search through other activity params,
        # then db params, then project params
        for new_group in Group.get(name=group).order:
            names = set()
            for name in ActivityParameter.static(new_group, only=needed):
                names.add(name)
                needed.remove(name)
            if names:
                chain.append({"kind": "activity", "group": new_group, "names": names})

        if needed and include_self:
            names = set()
            included = needed.intersection(data)
            for name in included:
                names.add(name)
                needed.remove(name)
            if names:
                chain.append({"kind": "activity", "group": group, "names": names})

        if needed:
            database = ActivityParameter.get(group=group).database
            names = set()
            for name in DatabaseParameter.static(database, only=needed):
                names.add(name)
                needed.remove(name)
            if names:
                chain.append({"kind": "database", "group": database, "names": names})
        if needed:
            names = set()
            for name in ProjectParameter.static(only=needed):
                names.add(name)
                needed.remove(name)
            if names:
                chain.append({"kind": "project", "group": "project", "names": names})
        if needed:
            raise MissingName(
                "The following variables aren't defined:\n{}".format("|".join(needed))
            )

        return chain

    @staticmethod
    def is_dependency_within_group(name, group, include_order=False):
        """ Determine if the given parameter `name` is a dependency within
        the given activity `group`.

        The optional ``include_order`` parameter will include dependencies
        from groups found in the the ``Group``.`order` field.
        """
        chain = ActivityParameter.dependency_chain(group, include_self=True)
        own_group = next((x for x in chain if x.get("group") == group), {})
        names = own_group.get("names", set())
        if include_order:
            for new_group in Group.get(name=group).order:
                order_group = next(
                    (x for x in chain if x.get("group") == new_group), {}
                )
                names = names.union(order_group.get("names", set()))
        return True if name in names else False

    @staticmethod
    def recalculate(group):
        """Recalculate all values for activity parameters in this group, and update their underlying `Activity` and `Exchange` values."""
        # Start by traversing and updating the list of dependencies
        if not ActivityParameter.expired(group):
            return

        chain = ActivityParameter.dependency_chain(group)

        # Reset dependencies and dependency order
        if chain:
            obj = Group.get(name=group)
            obj.order = [o["group"] for o in chain if o["kind"] == "activity"]
            obj.save()
            GroupDependency.delete().where(GroupDependency.group == group).execute()
            GroupDependency.insert_many(
                [{"group": group, "depends": o["group"]} for o in chain]
            ).execute()

        # Update all upstream groups
        mapping = {
            "project": ProjectParameter,
            "database": DatabaseParameter,
            "activity": ActivityParameter,
        }

        # Not guaranteed to be the most efficient,
        # but definitely simplest for now.
        # Could be smarter here in the future
        # Shouldn't be any race conditions because check for
        # circular dependencies
        for row in chain[::-1]:
            mapping[row["kind"]].recalculate(row["group"])

        # Update activity parameter values
        data = ActivityParameter.load(group)
        static = {
            k: v
            for k, v in ActivityParameter._static_dependencies(group).items()
            if k not in data
        }
        ParameterSet(data, static).evaluate_and_set_amount_field()
        with parameters.db.atomic():
            for key, value in data.items():
                ActivityParameter.update(amount=value["amount"],).where(
                    ActivityParameter.name == key, ActivityParameter.group == group,
                ).execute()
            Group.get(name=group).freshen()
            ActivityParameter.expire_downstream(group)

        ActivityParameter.recalculate_exchanges(group)

    @staticmethod
    def recalculate_exchanges(group):
        """Recalculate formulas for all parameterized exchanges in group ``group``."""
        if ActivityParameter.expired(group):
            return ActivityParameter.recalculate(group)

        interpreter = Interpreter()
        for k, v in ActivityParameter.static(group, full=True).items():
            interpreter.symtable[k] = v
        # TODO: Remove uncertainty from exchanges?
        for obj in ParameterizedExchange.select().where(
            ParameterizedExchange.group == group
        ):
            exc = ExchangeDataset.get(id=obj.exchange)
            exc.data["amount"] = interpreter(obj.formula)
            exc.save()

        databases.set_dirty(ActivityParameter.get(group=group).database)

    def save(self, *args, **kwargs):
        """Save this model instance"""
        Group.get_or_create(name=self.group)[0].expire()
        super(ActivityParameter, self).save(*args, **kwargs)

    def is_deletable(self):
        """Perform a test to see if the current parameter can be deleted."""
        # First check own group
        if ActivityParameter.is_dependency_within_group(self.name, self.group):
            return False
        # Then test other relevant activity groups.
        if ActivityParameter.is_dependent_on(self.name, self.group):
            return False
        return True

    @staticmethod
    def is_dependent_on(name, group):
        """ Test if any activity parameters are dependent on the given
        parameter name from the given group.
        """
        query = (
            GroupDependency.select(GroupDependency.group)
            .where(GroupDependency.depends == group)
            .distinct()
        )

        for row in query.execute():
            chain = ActivityParameter.dependency_chain(row.group)
            own_group = next((x for x in chain if x.get("group") == group), {})
            if name in own_group.get("names", set()):
                return True

        return False

    @classmethod
    def update_formula_project_parameter_name(cls, old, new):
        """ Performs an update of the formula of relevant parameters.

        This method specifically targets project parameters used in activity
        formulas
        """
        data = (
            alter_parameter_formula(p, old, new)
            for p in (
                cls.select()
                .join(GroupDependency, on=(GroupDependency.group == cls.group))
                .where(
                    (GroupDependency.depends == "project") & (cls.formula.contains(old))
                )
            )
            if not ActivityParameter.is_dependency_within_group(old, p.group)
        )
        group_parameters = itertools.chain(
            (
                cls.select(cls.group)
                .join(GroupDependency, on=(GroupDependency.group == cls.group))
                .where(
                    (GroupDependency.depends == "project") & (cls.formula.contains(old))
                )
                .distinct()
            ),
            (
                ParameterizedExchange.select(ParameterizedExchange.group)
                .where(ParameterizedExchange.formula.contains(old))
                .distinct()
            ),
        )
        groups = set(
            p.group
            for p in group_parameters
            if not ActivityParameter.is_dependency_within_group(old, p.group)
        )
        exchanges = (
            alter_parameter_formula(p, old, new)
            for p in ParameterizedExchange.select().where(
                ParameterizedExchange.group << groups
            )
        )
        cls.bulk_update(data, fields=[cls.formula], batch_size=50)
        for param_exc in exchanges:
            param_exc.save()
        Group.update(fresh=False).where(Group.name << groups).execute()

    @classmethod
    def update_formula_database_parameter_name(cls, old, new):
        """ Performs an update of the formula of relevant parameters.

        This method specifically targets database parameters used in activity
        formulas
        """
        data = (
            alter_parameter_formula(p, old, new)
            for p in (
                cls.select()
                .join(GroupDependency, on=(GroupDependency.group == cls.group))
                .where(
                    (GroupDependency.depends == cls.database)
                    & (cls.formula.contains(old))
                )
            )
            if not ActivityParameter.is_dependency_within_group(old, p.group)
        )
        group_parameters = itertools.chain(
            (
                cls.select(cls.group)
                .join(GroupDependency, on=(GroupDependency.group == cls.group))
                .where(
                    (GroupDependency.depends == cls.database)
                    & (cls.formula.contains(old))
                )
                .distinct()
            ),
            (
                ParameterizedExchange.select(ParameterizedExchange.group)
                .where(ParameterizedExchange.formula.contains(old))
                .distinct()
            ),
        )
        groups = set(
            p.group
            for p in group_parameters
            if not ActivityParameter.is_dependency_within_group(old, p.group)
        )
        exchanges = (
            alter_parameter_formula(p, old, new)
            for p in ParameterizedExchange.select().where(
                ParameterizedExchange.group << groups
            )
        )
        cls.bulk_update(data, fields=[cls.formula], batch_size=50)
        for param_exc in exchanges:
            param_exc.save()
        Group.update(fresh=False).where(Group.name << groups).execute()

    @classmethod
    def update_formula_activity_parameter_name(cls, old, new, include_order=False):
        """ Performs an update of the formula of relevant parameters.

        This method specifically targets activity parameters used in activity
        formulas
        """
        data = (
            alter_parameter_formula(p, old, new)
            for p in cls.select().where(cls.formula.contains(old))
            if ActivityParameter.is_dependency_within_group(old, p.group, include_order)
        )
        group_parameters = itertools.chain(
            cls.select(cls.group).where(cls.formula.contains(old)).distinct(),
            (
                ParameterizedExchange.select(ParameterizedExchange.group)
                .where(ParameterizedExchange.formula.contains(old))
                .distinct()
            ),
        )
        groups = set(
            p.group
            for p in group_parameters
            if ActivityParameter.is_dependency_within_group(old, p.group, include_order)
        )
        exchanges = (
            alter_parameter_formula(p, old, new)
            for p in ParameterizedExchange.select().where(
                ParameterizedExchange.group << groups
            )
        )
        cls.bulk_update(data, fields=[cls.formula], batch_size=50)
        for param_exc in exchanges:
            param_exc.save()
        Group.update(fresh=False).where(Group.name << groups).execute()

    @classmethod
    def create_table(cls):
        super(ActivityParameter, cls).create_table()
        cls._meta.database.execute_sql(CROSSDATASE_UPDATE_TRIGGER)
        cls._meta.database.execute_sql(CROSSDATASE_INSERT_TRIGGER)
        cls._meta.database.execute_sql(CROSSGROUP_UPDATE_TRIGGER)
        cls._meta.database.execute_sql(CROSSGROUP_INSERT_TRIGGER)

    @property
    def dict(self):
        """Parameter data as a standardized dictionary"""
        obj = nonempty(
            {
                "database": self.database,
                "code": self.code,
                "name": self.name,
                "formula": self.formula,
                "amount": self.amount,
            }
        )
        obj.update(self.data)
        return obj


class ParameterizedExchange(Model):
    group = TextField()
    exchange = IntegerField(unique=True)
    formula = TextField()

    @classmethod
    def create_table(cls):
        super(ParameterizedExchange, cls).create_table()
        cls._meta.database.execute_sql(PE_UPDATE_TRIGGER)
        cls._meta.database.execute_sql(PE_INSERT_TRIGGER)

    def save(self, *args, **kwargs):
        Group.get_or_create(name=self.group)[0].expire()
        super(ParameterizedExchange, self).save(*args, **kwargs)
        # Push the changed formula to the Exchange.
        exc = ExchangeDataset.get_or_none(id=self.exchange)
        if exc and exc.data.get("formula") != self.formula:
            exc.data["formula"] = self.formula
            exc.save()

    @staticmethod
    def load(group):
        """Return dictionary of parameter data with names as keys and ``.dict()`` as values."""
        return {
            o.exchange: o.formula
            for o in ParameterizedExchange.select().where(
                ParameterizedExchange.group == group
            )
        }

    @staticmethod
    def recalculate(group):
        """Shortcut for ``ActivityParameter.recalculate_exchanges``."""
        return ActivityParameter.recalculate_exchanges(group)


class Group(Model):
    name = TextField(unique=True)
    fresh = BooleanField(default=True)
    updated = DateTimeField(default=datetime.datetime.now)
    order = PickleField(default=[])

    def expire(self):
        """Set ``fresh`` to ``False``"""
        self.fresh = False
        self.save()

    def freshen(self):
        """Set ``fresh`` to ``True``"""
        self.fresh = True
        self.save()

    def save(self, *args, **kwargs):
        """Save this model instance. Will remove 'project' and database names from ``order``."""
        self.purge_order()
        super(Group, self).save(*args, **kwargs)

    def purge_order(self):
        reserved = set(databases).union(set(["project"]))
        self.order = [x for x in self.order if x not in reserved]

    class Meta:
        table_name = "group_table"


class GroupDependency(Model):
    group = TextField()
    depends = TextField()

    class Meta:
        indexes = ((("group", "depends"), True),)
        constraints = [Check('"group" != depends')]

    def save(self, *args, **kwargs):
        if self.group == "project":
            raise ValueError("`project` group can't have dependencies")
        elif self.group in databases and self.depends != "project":
            raise ValueError("Database groups can only depend on `project`")
        super(GroupDependency, self).save(*args, **kwargs)

    @classmethod
    def create_table(cls):
        super(GroupDependency, cls).create_table()
        cls._meta.database.execute_sql(GD_UPDATE_TRIGGER)
        cls._meta.database.execute_sql(GD_INSERT_TRIGGER)


class ParameterManager(object):
    def __init__(self):
        self.db = SubstitutableDatabase(
            projects.dir / "parameters.db",
            [
                DatabaseParameter,
                ProjectParameter,
                ActivityParameter,
                ParameterizedExchange,
                Group,
                GroupDependency,
            ],
        )
        config.sqlite3_databases.append(("parameters.db", self.db))

    def add_to_group(self, group, activity):
        """Add `activity` to group.

        Creates ``group`` if needed.

        Will delete any existing ``ActivityParameter`` for this activity.

        Deletes `parameters` key from `Activity`."""
        Group.get_or_create(name=group)

        activity = get_activity((activity[0], activity[1]))
        if "parameters" not in activity:
            return

        # Avoid duplicate by deleting existing parameters
        ActivityParameter.delete().where(
            ActivityParameter.database == activity["database"],
            ActivityParameter.code == activity["code"],
        ).execute()

        def reformat(o):
            skipped = ("name", "amount", "formula")
            return [
                nonempty(
                    {
                        "group": group,
                        "database": o["database"],
                        "code": o["code"],
                        "name": p["name"],
                        "formula": p.get("formula"),
                        "amount": p.get("amount", 0),
                        "data": {k: v for k, v in p.items() if k not in skipped},
                    }
                )
                for p in o.get("parameters", [])
            ]

        # Get formatted parameters
        with self.db.atomic():
            for row in reformat(activity):
                ActivityParameter.create(**row)

        # Parameters are now "active", remove from `Activity`
        del activity["parameters"]
        activity.save()

        self.add_exchanges_to_group(group, activity)

        return (
            ActivityParameter.select()
            .where(
                ActivityParameter.database == activity["database"],
                ActivityParameter.code == activity["code"],
            )
            .count()
        )

    def remove_from_group(self, group, activity, restore_amounts=True):
        """Remove `activity` from `group`.

        Will delete any existing ``ActivityParameter`` and ``ParameterizedExchange`` for this activity.

        Restores `parameters` key to this `Activity`.
        By default, restores `amount` value of each parameterized exchange
        of the `Activity` to the original value. This can be avoided by using
        the ``restore_amounts`` parameter.

        """

        def drop_fields(dct):
            dct = {k: v for k, v in dct.items() if k not in ("database", "code")}
            return dct.pop("name"), dct

        activity = get_activity((activity[0], activity[1]))
        activity["parameters"] = dict(
            [
                drop_fields(o.dict)
                for o in ActivityParameter.select().where(
                    ActivityParameter.database == activity[0],
                    ActivityParameter.code == activity[1],
                )
            ]
        )

        with self.db.atomic():
            self.remove_exchanges_from_group(group, activity, restore_amounts)
            ActivityParameter.delete().where(
                ActivityParameter.database == activity[0],
                ActivityParameter.code == activity[1],
            ).execute()
            activity.save()

    def add_exchanges_to_group(self, group, activity):
        """ Add exchanges with formulas from ``activity`` to ``group``.

        Every exchange with a formula field will have its original `amount`
        value stored as `original_amount`. This original value can be
        restored when parameterization is removed from the activity with
        `remove_from_group`.

        """
        count = 0
        if (
            not ActivityParameter.select()
            .where(
                ActivityParameter.database == activity[0],
                ActivityParameter.code == activity[1],
            )
            .count()
        ):
            ActivityParameter.insert_dummy(group, activity)

        for exc in get_activity((activity[0], activity[1])).exchanges():
            if "formula" in exc:
                try:
                    obj = ParameterizedExchange.get(exchange=exc._document.id)
                except ParameterizedExchange.DoesNotExist:
                    obj = ParameterizedExchange(exchange=exc._document.id)
                obj.group = group
                obj.formula = exc["formula"]
                obj.save()
                if "original_amount" not in exc:
                    exc["original_amount"] = exc["amount"]
                    exc.save()
                count += 1

        return count

    def remove_exchanges_from_group(self, group, activity, restore_original=True):
        """ Takes a group and activity and removes all ``ParameterizedExchange``
        objects from the group.

        The ``restore_original`` parameter determines if the original amount
        values will be restored to those exchanges where a formula was used
        to alter the amount.

        """
        if restore_original:
            for exc in (ex for ex in activity.exchanges() if "original_amount" in ex):
                exc["amount"] = exc["original_amount"]
                del exc["original_amount"]
                exc.save()

        ParameterizedExchange.delete().where(
            ParameterizedExchange.group == group
        ).execute()

    def new_project_parameters(self, data, overwrite=True):
        """Efficiently and correctly enter multiple parameters.

        Will overwrite existing project parameters with the same name, unless ``overwrite`` is false, in which case a ``ValueError`` is raised.

        ``data`` should be a list of dictionaries:

        .. code-block:: python

            [{
                'name': name of variable (unique),
                'amount': numeric value of variable (optional),
                'formula': formula in Python as string (optional),
                optional keys like uncertainty, etc. (no limitations)
            }]

        """
        potentially_non_unique_names = [ds["name"] for ds in data]
        unique_names = list(set(potentially_non_unique_names))
        assert len(unique_names) == len(
            potentially_non_unique_names
        ), "Nonunique names: {}".format(
            [p for p in unique_names if potentially_non_unique_names.count(p) > 1]
        )

        def reformat(ds):
            return {
                "name": ds.pop("name"),
                "amount": ds.pop("amount", 0),
                "formula": ds.pop("formula", None),
                "data": ds,
            }

        data = [reformat(ds) for ds in data]
        new = {o["name"] for o in data}
        existing = {
            o[0] for o in ProjectParameter.select(ProjectParameter.name).tuples()
        }

        if new.intersection(existing) and not overwrite:
            raise ValueError(
                "The following parameters already exist:\n{}".format(
                    "|".join(new.intersection(existing))
                )
            )

        with self.db.atomic():
            # Remove existing values
            ProjectParameter.delete().where(
                ProjectParameter.name << tuple(new)
            ).execute()
            for idx in range(0, len(data), 100):
                ProjectParameter.insert_many(data[idx : idx + 100]).execute()
            Group.get_or_create(name="project")[0].expire()
            ProjectParameter.recalculate()

    def new_database_parameters(self, data, database, overwrite=True):
        """Efficiently and correctly enter multiple parameters. Deletes **all** existing database parameters for this database.

        Will overwrite existing database parameters with the same name, unless ``overwrite`` is false, in which case a ``ValueError`` is raised.

        ``database`` should be an existing database. ``data`` should be a list of dictionaries:

        .. code-block:: python

            [{
                'name': name of variable (unique),
                'amount': numeric value of variable (optional),
                'formula': formula in Python as string (optional),
                optional keys like uncertainty, etc. (no limitations)
            }]

        """
        assert database in databases, "Unknown database"

        potentially_non_unique_names = [ds["name"] for ds in data]
        unique_names = list(set(potentially_non_unique_names))
        assert len(unique_names) == len(
            potentially_non_unique_names
        ), "Nonunique names: {}".format(
            [p for p in unique_names if potentially_non_unique_names.count(p) > 1]
        )

        def reformat(ds):
            return {
                "database": database,
                "name": ds.pop("name"),
                "amount": ds.pop("amount", 0),
                "formula": ds.pop("formula", None),
                "data": ds,
            }

        data = [reformat(ds) for ds in data]
        new = {o["name"] for o in data}
        existing = {
            o[0]
            for o in DatabaseParameter.select(DatabaseParameter.name)
            .where(DatabaseParameter.database == database)
            .tuples()
        }

        if new.intersection(existing) and not overwrite:
            raise ValueError(
                "The following parameters already exist:\n{}".format(
                    "|".join(new.intersection(existing))
                )
            )

        with self.db.atomic():
            # Remove existing values
            DatabaseParameter.delete().where(
                DatabaseParameter.database == database,
                DatabaseParameter.name << tuple(new),
            ).execute()
            for idx in range(0, len(data), 100):
                DatabaseParameter.insert_many(data[idx : idx + 100]).execute()
            Group.get_or_create(name=database)[0].expire()
            DatabaseParameter.recalculate(database)

    def new_activity_parameters(self, data, group, overwrite=True):
        """Efficiently and correctly enter multiple parameters. Deletes **all** existing activity parameters for this group.

        Will overwrite existing parameters in the same group with the same name, unless ``overwrite`` is false, in which case a ``ValueError`` is raised.

        Input parameters must refer to a single, existing database.

        ``group`` is the group name; will be autocreated if necessary. ``data`` should be a list of dictionaries:

        .. code-block:: python

            [{
                'name': name of variable (unique),
                'database': activity database,
                'code': activity code,
                'amount': numeric value of variable (optional),
                'formula': formula in Python as string (optional),
                optional keys like uncertainty, etc. (no limitations)
            }]

        """
        database = {o["database"] for o in data}
        assert len(database) == 1, "Multiple databases"
        assert database.pop() in databases, "Unknown database"

        potentially_non_unique_names = [o["name"] for o in data]
        unique_names = list(set(potentially_non_unique_names))
        assert len(unique_names) == len(
            potentially_non_unique_names
        ), "Nonunique names: {}".format(
            [p for p in unique_names if potentially_non_unique_names.count(p) > 1]
        )

        Group.get_or_create(name=group)

        def reformat(ds):
            return {
                "group": group,
                "database": ds.pop("database"),
                "code": ds.pop("code"),
                "name": ds.pop("name"),
                "formula": ds.pop("formula", None),
                "amount": ds.pop("amount", 0),
                "data": ds,
            }

        data = [reformat(ds) for ds in data]
        new = {o["name"] for o in data}
        existing = {
            o[0]
            for o in ActivityParameter.select(ActivityParameter.name)
            .where(ActivityParameter.group == group)
            .tuples()
        }

        if new.intersection(existing) and not overwrite:
            raise ValueError(
                "The following parameters already exist:\n{}".format(
                    "|".join(new.intersection(existing))
                )
            )

        with self.db.atomic():
            # Remove existing values
            ActivityParameter.delete().where(
                ActivityParameter.group == group, ActivityParameter.name << new
            ).execute()
            for idx in range(0, len(data), 100):
                ActivityParameter.insert_many(data[idx : idx + 100]).execute()
            Group.get_or_create(name=group)[0].expire()
            ActivityParameter.recalculate(group)

    def rename_project_parameter(self, parameter, new_name, update_dependencies=False):
        """ Given a parameter and a new name, safely update the parameter.

        Will raise a TypeError if the given parameter is of the incorrect type.
        Will raise a ValueError if other parameters depend on the given one
        and ``update_dependencies`` is False.

        """
        if not isinstance(parameter, ProjectParameter):
            raise TypeError("Incorrect parameter type for this method.")
        if parameter.name == new_name:
            return

        project = ProjectParameter.is_dependency_within_group(parameter.name)
        database = DatabaseParameter.is_dependent_on(parameter.name)
        activity = ActivityParameter.is_dependent_on(parameter.name, "project")

        if not update_dependencies and any([project, database, activity]):
            raise ValueError(
                "Parameter '{}' is used in other (downstream) formulas".format(
                    parameter.name
                )
            )

        with self.db.atomic():
            if project:
                ProjectParameter.update_formula_parameter_name(parameter.name, new_name)
            if database:
                DatabaseParameter.update_formula_project_parameter_name(
                    parameter.name, new_name
                )
            if activity:
                ActivityParameter.update_formula_project_parameter_name(
                    parameter.name, new_name
                )
            parameter.name = new_name
            parameter.save()
            self.recalculate()

    def rename_database_parameter(self, parameter, new_name, update_dependencies=False):
        """ Given a parameter and a new name, safely update the parameter.

        Will raise a TypeError if the given parameter is of the incorrect type.
        Will raise a ValueError if other parameters depend on the given one
        and ``update_dependencies`` is False.

        """
        if not isinstance(parameter, DatabaseParameter):
            raise TypeError("Incorrect parameter type for this method.")
        if parameter.name == new_name:
            return

        database = DatabaseParameter.is_dependency_within_group(
            parameter.name, parameter.database
        )
        activity = ActivityParameter.is_dependent_on(parameter.name, parameter.database)

        if not update_dependencies and any([database, activity]):
            raise ValueError(
                "Parameter '{}' is used in other (downstream) formulas".format(
                    parameter.name
                )
            )

        with self.db.atomic():
            if database:
                DatabaseParameter.update_formula_database_parameter_name(
                    parameter.name, new_name
                )
            if activity:
                ActivityParameter.update_formula_database_parameter_name(
                    parameter.name, new_name
                )
            parameter.name = new_name
            parameter.save()
            self.recalculate()

    def rename_activity_parameter(self, parameter, new_name, update_dependencies=False):
        """ Given a parameter and a new name, safely update the parameter.

        Will raise a TypeError if the given parameter is of the incorrect type.
        Will raise a ValueError if other parameters depend on the given one
        and ``update_dependencies`` is False.

        """
        if not isinstance(parameter, ActivityParameter):
            raise TypeError("Incorrect parameter type for this method.")
        if parameter.name == new_name:
            return

        activity = any(
            [
                ActivityParameter.is_dependency_within_group(
                    parameter.name, parameter.group, include_order=True
                ),
                ActivityParameter.is_dependent_on(parameter.name, parameter.group),
            ]
        )

        if not update_dependencies and activity:
            raise ValueError(
                "Parameter '{}' is used in other (downstream) formulas".format(
                    parameter.name
                )
            )

        with self.db.atomic():
            if activity:
                ActivityParameter.update_formula_activity_parameter_name(
                    parameter.name, new_name, include_order=True
                )
            parameter.name = new_name
            parameter.save()
            self.recalculate()

    def recalculate(self):
        """Recalculate all expired project, database, and activity parameters, as well as exchanges."""
        if ProjectParameter.expired():
            ProjectParameter.recalculate()
        for db in databases:
            if DatabaseParameter.expired(db):
                DatabaseParameter.recalculate(db)
        for obj in Group.select().where(Group.fresh == False):
            # Shouldn't be possible? Maybe concurrent access?
            if obj.name in databases or obj.name == "project":
                continue
            ActivityParameter.recalculate(obj.name)
            ActivityParameter.recalculate_exchanges(obj.name)

    def __len__(self):
        return (
            DatabaseParameter.select().count()
            + ProjectParameter.select().count()
            + ActivityParameter.select().count()
        )

    def __repr__(self):
        return "Parameters manager with {} objects".format(len(self))


parameters = ParameterManager()


def get_new_symbols(data, context=None):
    interpreter = asteval.Interpreter()
    BUILTIN_SYMBOLS = set(interpreter.symtable).union(set(context or set()))
    found = set()
    for ds in data:
        if isinstance(ds, str):
            formula = ds
        elif "formula" in ds:
            formula = ds["formula"]
        else:
            continue

        nf = asteval.NameFinder()
        nf.generic_visit(interpreter.parse(formula))
        found.update(set(nf.names))
    return found.difference(BUILTIN_SYMBOLS)


def alter_parameter_formula(parameter, old, new):
    """ Replace the `old` part with `new` in the formula field and return
    the parameter itself.
    """
    if hasattr(parameter, "formula"):
        parameter.formula = re.sub(r"\b{}\b".format(old), new, parameter.formula)
    return parameter



from bw2data.tests import bw2test
from bw2data import databases
from bw2data.search import Searcher, IndexManager
from bw2data.backends import SQLiteBackend


@bw2test
def test_search_dataset_containing_stop_word():
    im = IndexManager("foo")
    im.add_dataset(
        {"database": "foo", "code": "bar", "name": "foo of bar, high voltage"}
    )
    with Searcher("foo") as s:
        assert s.search("foo of bar, high voltage", proxy=False)


@bw2test
def test_add_dataset():
    im = IndexManager("foo")
    im.add_dataset({"database": "foo", "code": "bar", "name": "lollipop"})
    with Searcher("foo") as s:
        assert s.search("lollipop", proxy=False)


@bw2test
def test_search_dataset():
    im = IndexManager("foo")
    im.add_dataset({"database": "foo", "code": "bar", "name": "lollipop"})
    with Searcher("foo") as s:
        assert s.search("lollipop", proxy=False) == [
            {
                "comment": "",
                "product": "",
                "name": "lollipop",
                "database": "foo",
                "location": "",
                "code": "bar",
                "categories": "",
                "synonyms": "",
            }
        ]


@bw2test
def test_search_geocollection_location():
    im = IndexManager("foo")
    im.add_dataset(
        {
            "database": "foo",
            "code": "bar",
            "name": "lollipop",
            "location": ("foo", "Here"),
        }
    )
    with Searcher("foo") as s:
        assert s.search("lollipop", proxy=False) == [
            {
                "comment": "",
                "product": "",
                "name": "lollipop",
                "database": "foo",
                "location": "here",
                "code": "bar",
                "categories": "",
                "synonyms": "",
            }
        ]


@bw2test
def test_update_dataset():
    im = IndexManager("foo")
    ds = {"database": "foo", "code": "bar", "name": "lollipop"}
    im.add_dataset(ds)
    ds["name"] = "lemon cake"
    im.update_dataset(ds)
    with Searcher("foo") as s:
        assert s.search("lemon", proxy=False) == [
            {
                "comment": "",
                "product": "",
                "name": "lemon cake",
                "database": "foo",
                "location": "",
                "code": "bar",
                "categories": "",
                "synonyms": "",
            }
        ]


@bw2test
def test_delete_dataset():
    im = IndexManager("foo")
    ds = {"database": "foo", "code": "bar", "name": "lollipop"}
    im.add_dataset(ds)
    with Searcher("foo") as s:
        assert s.search("lollipop", proxy=False)
    im.delete_dataset(ds)
    with Searcher("foo") as s:
        assert not s.search("lollipop", proxy=False)


@bw2test
def test_add_datasets():
    im = IndexManager("foo")
    ds = [{"database": "foo", "code": "bar", "name": "lollipop"}]
    im.add_datasets(ds)
    with Searcher("foo") as s:
        assert s.search("lollipop", proxy=False)


@bw2test
def test_add_database():
    db = SQLiteBackend("foo")
    ds = {("foo", "bar"): {"database": "foo", "code": "bar", "name": "lollipop"}}
    db.write(ds)
    with Searcher(db.filename) as s:
        assert s.search("lollipop", proxy=False)
    db.make_unsearchable()
    with Searcher(db.filename) as s:
        assert not s.search("lollipop", proxy=False)


@bw2test
def test_add_searchable_database():
    db = SQLiteBackend("foo")
    ds = {("foo", "bar"): {"database": "foo", "code": "bar", "name": "lollipop"}}
    db.write(ds)
    with Searcher(db.filename) as s:
        assert s.search("lollipop", proxy=False)


@bw2test
def test_modify_database():
    db = SQLiteBackend("foo")
    ds = {("foo", "bar"): {"database": "foo", "code": "bar", "name": "lollipop"}}
    db.write(ds)
    with Searcher(db.filename) as s:
        assert not s.search("cream", proxy=False)
        assert s.search("lollipop", proxy=False)
    ds2 = {("foo", "bar"): {"database": "foo", "code": "bar", "name": "ice cream"}}
    db.write(ds2)
    with Searcher(db.filename) as s:
        assert s.search("cream", proxy=False)


@bw2test
def test_delete_database():
    db = SQLiteBackend("foo")
    ds = {("foo", "bar"): {"database": "foo", "code": "bar", "name": "lollipop"}}
    db.write(ds)
    with Searcher(db.filename) as s:
        assert s.search("lollipop", proxy=False)
    db.make_unsearchable()
    with Searcher(db.filename) as s:
        assert not s.search("lollipop", proxy=False)
    db.make_searchable()
    with Searcher(db.filename) as s:
        assert s.search("lollipop", proxy=False)
    del databases["foo"]
    with Searcher(db.filename) as s:
        assert not s.search("lollipop", proxy=False)


@bw2test
def test_reset_index():
    im = IndexManager("foo")
    ds = {"database": "foo", "code": "bar", "name": "lollipop"}
    im.add_dataset(ds)
    im.create()
    with Searcher("foo") as s:
        assert not s.search("lollipop", proxy=False)


@bw2test
def test_basic_search():
    im = IndexManager("foo")
    im.add_dataset({"database": "foo", "code": "bar", "name": "lollipop"})
    with Searcher("foo") as s:
        assert s.search("lollipop", proxy=False)


@bw2test
def test_product_term():
    im = IndexManager("foo")
    im.add_dataset({"database": "foo", "code": "bar", "reference product": "lollipop"})
    with Searcher("foo") as s:
        assert s.search("lollipop", proxy=False)


@bw2test
def test_comment_term():
    im = IndexManager("foo")
    im.add_dataset({"database": "foo", "code": "bar", "comment": "lollipop"})
    with Searcher("foo") as s:
        assert s.search("lollipop", proxy=False)


@bw2test
def test_categories_term():
    im = IndexManager("foo")
    im.add_dataset({"database": "foo", "code": "bar", "categories": ("lollipop",)})
    with Searcher("foo") as s:
        assert s.search("lollipop", proxy=False)


@bw2test
def test_limit():
    im = IndexManager("foo")
    im.add_datasets(
        [
            {"database": "foo", "code": "bar", "name": "lollipop {}".format(x)}
            for x in range(50)
        ]
    )
    with Searcher("foo") as s:
        assert len(s.search("lollipop", limit=25, proxy=False)) == 25


@bw2test
def test_search_faceting():
    im = IndexManager("foo")
    ds = [
        {"database": "foo", "code": "bar", "name": "lollipop", "location": "CH"},
        {"database": "foo", "code": "bar", "name": "ice lollipop", "location": "FR"},
    ]
    im.add_datasets(ds)
    with Searcher("foo") as s:
        res = s.search("lollipop", proxy=False, facet="location")
    assert res == {
        "fr": [
            {
                "comment": "",
                "product": "",
                "name": "ice lollipop",
                "database": "foo",
                "location": "fr",
                "code": "bar",
                "categories": "",
                "synonyms": "",
            }
        ],
        "ch": [
            {
                "comment": "",
                "product": "",
                "name": "lollipop",
                "database": "foo",
                "location": "ch",
                "code": "bar",
                "categories": "",
                "synonyms": "",
            }
        ],
    }


@bw2test
def test_copy_save_propogates_to_search_index():
    db = SQLiteBackend("foo")
    ds = {("foo", "bar"): {"database": "foo", "code": "bar", "name": "lollipop"}}
    db.write(ds)
    assert db.search("lollipop")
    cp = db.get("bar").copy(code="baz")
    cp["name"] = "candy"
    cp.save()
    assert db.search("candy")


@bw2test
def test_case_sensitivity_convert_lowercase():
    db = SQLiteBackend("foo")
    ds = {("foo", "bar"): {"database": "foo", "code": "bar", "name": "LOLLIpop"}}
    db.write(ds)
    assert db.search("LOLLIpop".lower())
    assert db.search("lollipop")
    assert db.search("LOLLipop")
    assert db.search("LOLL*")
    assert db.search("Lollipop")
    assert not db.search("nope")


@bw2test
def test_case_sensitivity_filter():
    db = SQLiteBackend("foo")
    ds = {
        ("foo", "bar"): {
            "database": "foo",
            "code": "bar",
            "name": "lollipop",
            "location": "CH",
        },
        ("foo", "baz"): {
            "database": "foo",
            "code": "baz",
            "name": "ice lollipop",
            "location": "FR",
        },
    }
    db.write(ds)
    assert len(db.search("lollipop")) == 2
    assert len(db.search("lollipop", filter={"location": "fr"})) == 1
    assert len(db.search("lollipop", filter={"location": "FR"})) == 1


@bw2test
def test_case_sensitivity_mask():
    db = SQLiteBackend("foo")
    ds = {
        ("foo", "bar"): {
            "database": "foo",
            "code": "bar",
            "name": "lollipop",
            "location": "CH",
        },
        ("foo", "baz"): {
            "database": "foo",
            "code": "baz",
            "name": "ice lollipop",
            "location": "FR",
            "reference product": "ZEBRA",
        },
    }
    db.write(ds)
    assert len(db.search("lollipop")) == 2
    assert len(db.search("lollipop", mask={"product": "ZEbra"})) == 1


@bw2test
def test_synonym_search():
    im = IndexManager("foo")
    im.add_dataset({"database": "foo", "code": "bar", "name": "polytetrafluoroethylene", "synonyms":["PTFE", "Teflon"]})
    with Searcher("foo") as s:
        assert s.search("Teflon", proxy=False) == [
            {
                "comment": "",
                "product": "",
                "name": "polytetrafluoroethylene",
                "database": "foo",
                "location": "",
                "code": "bar",
                "categories": "",
                "synonyms": "ptfe, teflon",
            }
        ]

if __name__ == "__main__":
    test_synonym_search()